#!/usr/bin/env python
"""Implementation of Proximal Policy Optimization Algorithm."""
import collections
import gc
import json
import os
import random
import re
import statistics
import time
from collections import OrderedDict
from glob import glob
from json import JSONDecodeError
from typing import Any
from typing import Callable
from typing import Dict
from typing import Tuple
from typing import Union

import gymnasium as gym

import numpy as np
import nvidia_smi
import psutil
import tensorflow as tf
from gymnasium.spaces import Box
from gymnasium.spaces import Discrete
from gymnasium.spaces import MultiDiscrete
from psutil import NoSuchProcess
from tqdm import tqdm

from angorapy import models
from angorapy.agent.dataio import read_dataset_from_storage
from angorapy.agent.gather import evaluate
from angorapy.agent.gather import Gatherer
from angorapy.agent.ppo.train_step import ff_train_step
from angorapy.agent.ppo.train_step import recurrent_train_step
from angorapy.common import const
from angorapy.common import policies
from angorapy.common.const import BASE_SAVE_PATH
from angorapy.common.const import COLORS
from angorapy.common.const import MIN_STAT_EPS
from angorapy.common.const import PATH_TO_EXPERIMENTS
from angorapy.common.const import PRETRAINED_COMPONENTS_PATH
from angorapy.common.const import STORAGE_DIR
from angorapy.common.mpi_optim import MpiAdam
from angorapy.common.policies import BasePolicyDistribution
from angorapy.common.policies import CategoricalPolicyDistribution
from angorapy.common.policies import GaussianPolicyDistribution
from angorapy.common.senses import Sensation
from angorapy.common.postprocessors import BaseRunningMeanPostProcessor
from angorapy.common.postprocessors import postprocessors_from_serializations
from angorapy.tasks.registration import make_task
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.utilities.core import env_extract_dims
from angorapy.utilities.core import find_optimal_tile_shape
from angorapy.utilities.core import flatten
from angorapy.utilities.core import mpi_flat_print
from angorapy.utilities.core import mpi_print
from angorapy.utilities.datatypes import condense_stats
from angorapy.utilities.datatypes import mpi_condense_stats
from angorapy.utilities.datatypes import StatBundle
from angorapy.utilities.error import ComponentError
from angorapy.utilities.model_utils import get_component
from angorapy.utilities.model_utils import get_layer_names
from angorapy.utilities.model_utils import is_recurrent_model
from angorapy.utilities.model_utils import requires_batch_size
from angorapy.utilities.model_utils import requires_sequence_length
from angorapy.utilities.model_utils import validate_env_model_compatibility
from angorapy.utilities.model_utils import validate_model_builder
from angorapy.utilities.statistics import ignore_none

try:
    from mpi4py import MPI
except:
    MPI = None


class PPOAgent:
    """Agent using the Proximal Policy Optimization Algorithm for learning.

    The default is an implementation using two independent models for the critic and the actor. This is of course more
    expensive than using shared parameters because we need two forward and backward calculations
    per batch however this is what is used in the original paper and most implementations. During development this also
    turned out to be beneficial for performance relative to episodes seen in easy tasks (e.g. CartPole) and crucial
    to make any significant progress in more difficult envs such as LunarLander.
    """

    policy: tf.keras.Model
    value: tf.keras.Model
    joint: tf.keras.Model

    def __init__(self, model_builder: Callable[..., Tuple[tf.keras.Model, tf.keras.Model, tf.keras.Model]],
                 environment: TaskWrapper, horizon: int = 1024, workers: int = 8, learning_rate: float = 0.001,
                 discount: float = 0.99, lam: float = 0.95, clip: float = 0.2, c_entropy: float = 0.01,
                 c_value: float = 0.5, gradient_clipping: float = None, clip_values: bool = True,
                 tbptt_length: int = 16, lr_schedule: str = None, distribution: BasePolicyDistribution = None,
                 reward_configuration: str = None, _make_dirs=True, debug: bool = False,
                 pretrained_components: list = None, n_optimizers: int = None):
        """ Initialize the PPOAgent with given hyperparameters. Policy and value network will be freshly initialized.

        Agent using the Proximal Policy Optimization Algorithm for learning.

        The default is an implementation using two independent models for the critic and the actor. This is of course more
        expensive than using shared parameters because we need two forward and backward calculations
        per batch however this is what is used in the original paper and most implementations. During development this also
        turned out to be beneficial for performance relative to episodes seen in easy tasks (e.g. CartPole) and crucial
        to make any significant progress in more difficult environments such as LunarLander.

        Args:
            model_builder: a function creating a policy, value and joint model
            environment (gym.Env): the environment in which the agent will learn
            horizon (int): the number of timesteps each worker collects
            workers (int): the number of workers
            learning_rate (float): the learning rate of the Adam optimizer
            discount (float): discount factor for future rewards during collection
            lam (float): lambda parameter of the generalized advantage estimation
            clip (float): clipping range for both policy and value loss
            c_entropy (float): coefficient for entropy in the combined loss
            c_value (float): coefficient fot value in the combined loss
            gradient_clipping (float): max norm for the gradient clipping, set None to deactivate (default)
            clip_values (bool): boolean switch to turn off or on the clipping of the value loss
            lr_schedule (str): type of scheduler for the learning rate, default None (constant learning rate). Can be
                either None or 'exponential'
            _make_dirs (bool): internal parameter to indicate whether to recreate the directories
            debug (bool): turn on/off debugging mode
        """
        super().__init__()
        self.debug = debug

        # setup distributed computation
        self.gpus = tf.config.list_physical_devices('GPU')
        if MPI is not None:
            self.mpi_comm = MPI.COMM_WORLD
            self.is_root = self.mpi_comm.rank == 0
            self.comm_size = self.mpi_comm.size
            self.comm_rank = self.mpi_comm.rank

            self.optimization_comm, self.is_optimization_process = self.get_optimization_comm(
                limit_to_n_optimizers=n_optimizers)
            self.n_optimizers = int(self.mpi_comm.allreduce(self.is_optimization_process, op=MPI.SUM))
        else:
            self.mpi_comm = None
            self.comm_size = 1
            self.comm_rank = 0
            self.is_root = True
            self.optimization_comm = None
            self.is_optimization_process = True
            self.n_optimizers = 1

        # checkups
        assert lr_schedule is None or isinstance(lr_schedule, str)

        # environment info
        self.env: TaskWrapper = environment
        self.env_name = self.env.unwrapped.spec.id
        self.state_dim, self.n_actions = env_extract_dims(self.env)
        if isinstance(self.env.action_space, (Discrete, MultiDiscrete)):
            self.continuous_control = False
        elif isinstance(self.env.action_space, Box):
            self.continuous_control = True
        else:
            raise NotImplementedError(f"PPO cannot handle unknown Action Space Typ: {self.env.action_space}")

        self.env.warmup()

        mpi_print(f"Using {self.env.transformers} for preprocessing.")

        # hyperparameters
        self.horizon = horizon
        self.n_workers = workers
        self.discount = discount
        self.learning_rate = learning_rate
        self.clip = tf.constant(clip, dtype=tf.float32)
        self.c_entropy = tf.constant(c_entropy, dtype=tf.float32)
        self.c_value = tf.constant(c_value, dtype=tf.float32)
        self.lam = lam
        self.gradient_clipping = gradient_clipping if gradient_clipping != 0 else None
        self.clip_values = clip_values
        self.tbptt_length = tbptt_length
        self.reward_configuration = reward_configuration

        # learning rate schedule
        self.lr_schedule_type = lr_schedule
        if lr_schedule is None:
            # adjust learning rate based on number of parallel GPUs
            self.lr_schedule = self.learning_rate
        elif lr_schedule.lower() == "exponential":
            self.lr_schedule = tf.keras.optimizers.schedules.ExponentialDecay(initial_learning_rate=self.learning_rate,
                                                                              decay_steps=workers * horizon,
                                                                              # decay after every cycle
                                                                              decay_rate=0.98)
        else:
            raise ValueError("Unknown Schedule type. Choose one of (None, exponential)")

        # models and optimizers
        self.distribution = distribution
        if self.distribution is None:
            self.distribution = CategoricalPolicyDistribution(
                self.env) if not self.continuous_control else GaussianPolicyDistribution(self.env)
        assert self.continuous_control == self.distribution.is_continuous, "Invalid distribution for environment."
        self.model_builder = model_builder
        self.builder_function_name = model_builder.__name__
        self.policy, self.value, self.joint = model_builder(env=self.env, distribution=self.distribution,
                                                            **({"bs": 1} if requires_batch_size(model_builder) else {}),
                                                            **({
                                                                   "sequence_length": self.tbptt_length} if requires_sequence_length(
                                                                model_builder) else {}))

        validate_model_builder(self.model_builder, self.env, self.distribution)
        validate_env_model_compatibility(self.env, self.joint)

        if pretrained_components is not None:
            print("Loading pretrained components:")
            for pretraining_name in pretrained_components:
                component_path = os.path.join(PRETRAINED_COMPONENTS_PATH, f"{pretraining_name}.h5")
                if os.path.isfile(component_path):
                    component = tf.keras.models.load_model(component_path, compile=False)
                else:
                    print(f"\tNo such pretraining found at {component_path}. Skipping.")
                    continue

                if component.name in get_layer_names(self.joint):
                    try:
                        get_component(self.joint, component.name).set_weights(component.get_weights())
                        print(f"\tSuccessfully loaded component '{component.name}'")
                    except ValueError as e:
                        print(f"Could not load weights into component: {e}")
                else:
                    print(f"\tNo outer component named '{component.name}' in model. Skipping.")

        self.optimizer: MpiAdam = MpiAdam(comm=self.optimization_comm, learning_rate=self.lr_schedule, epsilon=1e-5)
        self.optimizer.apply_gradients(
            zip([tf.zeros_like(v) for v in self.joint.trainable_variables], self.joint.trainable_variables))

        self.is_recurrent = is_recurrent_model(self.policy)
        if not self.is_recurrent:
            self.tbptt_length = 1

        self.gatherer_class = Gatherer

        # miscellaneous
        self.iteration = 0
        self.current_fps = 0
        self.gathering_fps = 0
        self.optimization_fps = 0
        self.device = "CPU:0"
        self.model_export_dir = "storage/saved_models/exports/"
        self.agent_id = f"{round(time.time())}{random.randint(int(1e5), int(1e6) - 1)}"
        if MPI is not None:
            self.agent_id = self.mpi_comm.bcast(self.agent_id, root=0)

        self.agent_directory = f"{BASE_SAVE_PATH}/{self.agent_id}/"
        self.experiment_directory = f"{PATH_TO_EXPERIMENTS}/{self.agent_id}/"
        if _make_dirs:
            os.makedirs(self.model_export_dir, exist_ok=True)
            os.makedirs(self.agent_directory, exist_ok=True)

        if os.path.isdir("storage/experience"):
            # TODO loading a currently training agent will potentially delete its storage
            # shutil.rmtree("storage/experience/")
            pass
        os.makedirs("storage/experience/", exist_ok=True)

        # statistics
        self.total_frames_seen = 0
        self.total_episodes_seen = 0
        self.years_of_experience = 0
        self.episode_reward_history = []
        self.episode_length_history = []
        self.cycle_reward_history = []
        self.cycle_length_history = []
        self.cycle_reward_std_history = []
        self.cycle_length_std_history = []
        self.cycle_stat_n_history = []
        self.entropy_history = []
        self.policy_loss_history = []
        self.value_loss_history = []
        self.time_dicts = []
        self.underflow_history = []
        self.loading_history = []
        self.current_per_receptor_mean = {}
        self.auxiliary_performances = {}

        # training statistics
        self.cycle_timings = []
        self.optimization_timings = []
        self.gathering_timings = []
        self.used_memory = []
        self.used_gpu_memory = []

        self.wrapper_stat_history = {}
        for transformer in self.env.transformers:
            self.wrapper_stat_history.update({transformer.__class__.__name__: {"mean": [transformer.simplified_mean()],
                                                                               "stdev": [
                                                                                   transformer.simplified_stdev()]}})

    @staticmethod
    def get_optimization_comm(limit_to_n_optimizers: int = None):
        mpi_comm = MPI.COMM_WORLD
        gpus = tf.config.list_physical_devices('GPU')
        is_root = mpi_comm.rank == 0

        if len(gpus) > 0:
            node_names = list(set(mpi_comm.allgather(MPI.Get_processor_name())))
            node_comm = mpi_comm.Split(color=node_names.index(MPI.Get_processor_name()))
            node_optimizer_rank = node_comm.allreduce(mpi_comm.rank, op=MPI.MIN)

            # TODO make this work for multiple GPUs per node?
            is_optimization_process = MPI.COMM_WORLD.rank == node_optimizer_rank
        else:
            is_optimization_process = True

        if not is_optimization_process:
            try:
                if tf.config.get_visible_devices("GPU"):
                    tf.config.set_visible_devices([], "GPU")
                os.environ["CUDA_VISIBLE_DEVICES"] = "-1"
            except:
                pass
        else:
            try:
                if len(gpus) > 0:
                    tf.config.experimental.set_memory_growth(gpus[0], True)
            except:
                pass

        optimization_ranks = [r for r in mpi_comm.allgather(MPI.COMM_WORLD.rank if is_optimization_process else -1) if
                              r != -1]
        if limit_to_n_optimizers is not None and limit_to_n_optimizers != 0:
            optimization_ranks = optimization_ranks[:limit_to_n_optimizers]
            is_optimization_process = mpi_comm.rank in optimization_ranks

        optimization_comm = mpi_comm.Split(color=(1 if is_optimization_process else 0))

        return optimization_comm, is_optimization_process

    def record_wrapper_stats(self) -> None:
        """Records the stats from RunningMeanWrappers."""
        for transformer in self.env.transformers:
            if transformer.name not in self.wrapper_stat_history.keys() or not isinstance(transformer,
                                                                                          BaseRunningMeanPostProcessor):
                continue

            self.wrapper_stat_history[transformer.__class__.__name__]["mean"].append(transformer.simplified_mean())
            self.wrapper_stat_history[transformer.__class__.__name__]["stdev"].append(transformer.simplified_stdev())

    def __repr__(self):
        return f"PPOAgent[at {self.iteration}][{self.env_name}]"

    def set_gpu(self, activated: bool) -> None:
        """Set GPU usage mode."""
        self.device = "GPU:0" if activated else "CPU:0"
        pass

    def assign_gatherer(self, new_gathering_class: Callable):
        self.gatherer_class = new_gathering_class

    def act(self, state: Union[Sensation, Dict[str, Any]]):
        """Sample an action from the agent's policy based on a given state. The sampled action is returned in a format
        that can be directly given to an environment.

        This method is mostly useful at inference time and serves as q quick wrapper around the steps required to
        process the raw numpy states of the environment into a state readable by the policy network, followed by
        sampling from the predicted distribution."""

        # based on given state, predict action distribution and state value; need flatten due to tf eager bug
        if isinstance(state, (dict, collections.OrderedDict)) and "observation" in state.keys():
            state = state["observation"]

            if not isinstance(state, Sensation):
                try:
                    state = Sensation(**state)
                except:
                    raise ValueError("Observation in state dict must be a Sensation or dict convertible into "
                                     "an observation.")
        elif isinstance(state, Sensation):
            pass
        else:
            raise ValueError("State must be a Sensation or a dictionary with an 'observation' key.")

        _, _, joint = self.build_models(self.joint.get_weights(), 1, 1)

        prepared_state = state.with_leading_dims(time=self.is_recurrent).dict_as_tf()
        policy_out = flatten(joint(prepared_state, training=False))

        predicted_distribution_parameters, value = policy_out[:-1], policy_out[-1]
        # from the action distribution sample an action and remember both the action and its probability
        action, action_probability = self.distribution.act(*predicted_distribution_parameters)

        return action

    def drill(self, n: int, epochs: int, batch_size: int, monitor: "Monitor" = None, save_every: int = 0,
              separate_eval: bool = False, stop_early: bool = False, radical_evaluation: object = False) -> "PPOAgent":
        """Start a training loop of the agent.

        Runs **n** cycles of experience gathering and optimization based on the gathered experience.

        Args:
            n (int): the number of experience-optimization cycles that shall be run
            epochs (int): the number of epochs for which the model is optimized on the same experience data
            batch_size (int): batch size for the optimization
            monitor: story telling object that creates visualizations of the training process on the fly (Default
                value = None)
            save_every (int): for any int x > 0 save the policy every x iterations, if x = 0 (default) do not save
            separate_eval (bool): if false (default), use episodes from gathering for statistics, if true, evaluate 10
                additional episodes.
            stop_early (bool): if true, stop the drill early if at least the previous 5 cycles achieved a performance
                above the envs threshold


        Returns:
            self, for chaining
        """

        # start monitor
        if self.is_root and monitor is not None:
            monitor.make_metadata(additional_hps={"epochs_per_cycle": epochs, "batch_size": batch_size, })

        self.batch_size = batch_size

        # determine the number of independent repeated gatherings required on this worker
        worker_base, worker_extra = divmod(self.n_workers, self.comm_size)
        worker_split = [worker_base + (r < worker_extra) for r in range(self.comm_size)]
        worker_collection_ids = list(range(self.n_workers))[
                                sum(worker_split[:self.comm_rank]):sum(worker_split[:self.comm_rank + 1])]

        # determine the split of worker outputs over optimizers
        optimizer_base, optimizer_extra = divmod(self.n_workers, self.n_optimizers)
        optimizer_split = [optimizer_base + (r < optimizer_extra) for r in range(self.n_optimizers)]
        optimizer_collection_ids = list(range(self.n_workers))[sum(optimizer_split[:self.optimizer.comm_rank]):sum(
            optimizer_split[:self.optimizer.comm_rank + 1])]

        if MPI is not None:
            list_of_worker_collection_id_lists = self.mpi_comm.gather(worker_collection_ids, root=0)
            list_of_optimizer_collection_id_lists = self.optimizer.comm.gather(optimizer_collection_ids, root=0)
        else:
            list_of_worker_collection_id_lists = [worker_collection_ids]
            list_of_optimizer_collection_id_lists = [optimizer_collection_ids]

        if self.is_root:
            print(f"\n\nDrill started using {self.comm_size} processes for {self.n_workers} workers of which "
                  f"{self.n_optimizers} are optimizers."
                  f" Worker distribution: {[worker_base + (r < worker_extra) for r in range(self.comm_size)]}.\n"
                  f"IDs over Workers: {list_of_worker_collection_id_lists}\n"
                  f"IDs over Optimizers: {list_of_optimizer_collection_id_lists}")

        if self.is_recurrent:
            assert batch_size % self.tbptt_length == 0, f"Batch size (the number of transitions per update)" \
                                                        f" must be a multiple of the sequence length "

            n_chunks_per_trajectory = self.horizon // self.tbptt_length
            n_chunks_per_batch = batch_size // self.tbptt_length
            n_chunks_per_batch_per_process = n_chunks_per_batch // self.n_optimizers
            n_trajectories_per_process = self.n_workers // self.n_optimizers

            n_trajectories_per_batch_per_process, n_chunks_per_trajectory_per_batch_per_process = find_optimal_tile_shape(
                (n_trajectories_per_process, n_chunks_per_trajectory), n_chunks_per_batch_per_process, width_first=True
                # TODO smartly adapt this to memory reqs or at least make parameter
            )

            n_trajectories_per_batch = n_trajectories_per_batch_per_process * self.n_optimizers
            n_chunks_per_trajectory_per_batch = n_chunks_per_trajectory_per_batch_per_process
            self.n_updates_per_epoch = (self.n_workers * self.horizon) // batch_size

            if self.is_root:
                print(f"\nThe policy is recurrent and the batch size is interpreted as the number of transitions "
                      f"per policy update. Given the batch size of {batch_size} this results in: \n"
                      f"\t{n_chunks_per_batch} chunks per update and {self.n_updates_per_epoch} updates per epoch\n"
                      f"\tBatch tilings of "
                      f"{n_trajectories_per_batch_per_process, n_chunks_per_trajectory_per_batch_per_process} per process "
                      f"and {n_trajectories_per_batch, n_chunks_per_trajectory_per_batch} in total.\n\n")

            batch_size = n_trajectories_per_batch_per_process
            effective_batch_size = (n_trajectories_per_batch_per_process, n_chunks_per_trajectory_per_batch_per_process)

        else:
            assert self.horizon * self.n_workers >= batch_size, "Batch Size is larger than the number of transitions."
            batch_size = batch_size // self.n_optimizers
            effective_batch_size = batch_size

        # rebuild model with desired batch size
        joint_weights = self.joint.get_weights()
        self.policy, self.value, self.joint = self.build_models(weights=joint_weights, batch_size=batch_size,
                                                                sequence_length=self.tbptt_length)

        _, _, actor_joint = self.build_models(joint_weights, 1, 1)

        # reset optimizer to not be confused by the newly build models
        # todo maybe its better to maintain a class attribute with optimizer weights and always write it in here
        optimizer_weights = self.optimizer.get_weights()
        if len(optimizer_weights) > 0:
            self.optimizer = MpiAdam(self.optimizer.comm, self.optimizer.learning_rate, self.optimizer.epsilon)
            self.optimizer.apply_gradients(
                zip([tf.zeros_like(v) for v in self.joint.trainable_variables], self.joint.trainable_variables))
            self.optimizer.set_weights(optimizer_weights)

        actor = self._make_actor(horizon=self.horizon, discount=self.discount, lam=self.lam,
                                 subseq_length=self.tbptt_length)

        cycle_start = None
        full_drill_start_time = time.time()
        for self.iteration in range(self.iteration, n):
            mpi_flat_print(f"Gathering cycle {self.iteration}...")

            time_dict = OrderedDict()
            subprocess_start = time.time()

            # distribute parameters from rank 0 to all goal ranks
            joint_weights = self.mpi_comm.bcast(joint_weights, root=0) if MPI is not None else joint_weights
            self.joint.set_weights(joint_weights)
            actor_joint.set_weights(joint_weights)

            # run simulations in parallel
            worker_stats = []

            for i in worker_collection_ids:
                stats = actor.collect(actor_joint, self.env, collector_id=i)
                worker_stats.append(stats)

            # merge gatherings from all workers
            stats = mpi_condense_stats(worker_stats)

            if MPI is not None:
                stats = self.mpi_comm.bcast(stats, root=0)

            # sync the envs to share statistics for transformers etc.
            self.env.mpi_sync()

            time_dict["gathering"] = time.time() - subprocess_start
            subprocess_start = time.time()

            # make seperate evaluation if necessary and wanted
            stats_with_evaluation = stats
            if separate_eval:
                if radical_evaluation or stats.numb_completed_episodes < MIN_STAT_EPS:
                    mpi_flat_print("Evaluating...")
                    n_evaluations = MIN_STAT_EPS if radical_evaluation else MIN_STAT_EPS - stats.numb_completed_episodes
                    evaluation_stats, _ = self.evaluate(n_evaluations)

                    if radical_evaluation:
                        stats_with_evaluation = evaluation_stats
                    else:
                        stats_with_evaluation = mpi_condense_stats([stats, evaluation_stats])
            elif self.comm_rank == 0 and stats.numb_completed_episodes == 0:
                print("WARNING: You are using a horizon that caused this cycle to not finish a single episode. "
                      "Consider activating separate evaluation in drill() to get meaningful statistics.")

            # RECORD STATS, SAVE MODEL AND REPORT
            if self.is_root:
                self.record_wrapper_stats()
                self.record_stats(stats_with_evaluation)

                time_dict["evaluating"] = time.time() - subprocess_start

                # save if best
                last_score = self.cycle_reward_history[-1]
                if last_score is not None and last_score == ignore_none(max, self.cycle_reward_history):
                    self.save_agent_state(name="best")

                if cycle_start is not None:
                    self.cycle_timings.append(round(time.time() - cycle_start, 2))

                used_memory = 0
                for pid in psutil.pids():
                    try:
                        p = psutil.Process(pid)
                        if "python" in p.name():
                            used_memory += p.memory_info()[0]
                    except NoSuchProcess:
                        pass

                used_gpu_memory = 0
                if len(self.gpus) > 0:
                    nvidia_smi.nvmlInit()
                    nvidia_handle = nvidia_smi.nvmlDeviceGetHandleByIndex(0)
                    procs = nvidia_smi.nvmlDeviceGetComputeRunningProcesses(nvidia_handle)
                    used_gpu_memory = 0
                    for p in procs:
                        try:
                            if "python" in psutil.Process(p.pid).name():
                                used_gpu_memory += p.usedGpuMemory
                        except NoSuchProcess:
                            pass

                self.used_memory.append(round(used_memory / 1e9, 2))
                self.used_gpu_memory.append(round(used_gpu_memory / 1e9, 2))
                self.report(total_iterations=n)
                cycle_start = time.time()

            # EARLY STOPPING  TODO check how this goes with MPI
            if self.env.spec.reward_threshold is not None and stop_early:
                if np.all(np.greater_equal(self.cycle_reward_history[-5:], self.env.spec.reward_threshold)):
                    print("\rAll catch a breath, we stop the drill early due to the formidable result!")
                    break

            # OPTIMIZATION PHASE
            if self.is_optimization_process:
                dataset = read_dataset_from_storage(dtype_actions=tf.float32 if self.continuous_control else tf.int32,
                                                    id_prefix=self.agent_id, worker_ids=optimizer_collection_ids,
                                                    responsive_senses=self.policy.input_names)

                subprocess_start = time.time()
                self.optimize(dataset, epochs, batch_size, effective_batch_size)
                time_dict["optimizing"] = time.time() - subprocess_start

                # FINALIZE
                if self.is_root:
                    mpi_flat_print("Finalizing...")
                    self.total_frames_seen += stats.numb_processed_frames
                    self.total_episodes_seen += stats.numb_completed_episodes
                    if hasattr(self.env, "dt"):
                        self.years_of_experience += (stats.numb_processed_frames * self.env.dt / 3.154e+7)
                    else:
                        self.years_of_experience = None

                    # update monitor logs
                    if monitor is not None:
                        if monitor.gif_every != 0 and (self.iteration + 1) % monitor.gif_every == 0:
                            print("Creating Episode GIFs for current state of policy...")
                            monitor.create_episode_gif(n=1)

                        if monitor.frequency != 0 and (self.iteration + 1) % monitor.frequency == 0:
                            monitor.update()

                    # save the current state of the agent
                    if save_every != 0 and self.iteration != 0 and (self.iteration + 1) % save_every == 0:
                        # every x iterations
                        self.save_agent_state()

                    # and (overwrite) the latest version
                    self.save_agent_state("last")

                    # calculate processing speed in fps
                    self.current_fps = stats.numb_processed_frames / (
                        sum([v for v in time_dict.values() if v is not None]))
                    self.gathering_fps = (stats.numb_processed_frames // min(self.n_workers, self.comm_size)) / (
                        time_dict["gathering"])
                    self.optimization_fps = (stats.numb_processed_frames * epochs) / (time_dict["optimizing"])
                    self.time_dicts.append(time_dict)

                    self.optimization_timings.append(time_dict["optimizing"])
                    self.gathering_timings.append(time_dict["gathering"])

                joint_weights = self.joint.get_weights()

            gc.collect()
            tf.compat.v1.reset_default_graph()
            tf.keras.backend.clear_session()

        # after training rebuild the network so that it is available to the agent
        self.policy, self.value, self.joint = self.build_models(weights=joint_weights, batch_size=batch_size,
                                                                sequence_length=self.tbptt_length)

        if self.is_root:
            print(f"Drill finished after {round(time.time() - full_drill_start_time, 2)}serialization.")

        return self

    def record_stats(self, stats):
        """Record a given StatsBundle in the history of the agent."""
        try:
            mean_eps_length = statistics.mean(stats.episode_lengths) if len(stats.episode_lengths) > 1 else \
                stats.episode_lengths[0]
            mean_eps_rewards = statistics.mean(stats.episode_rewards) if len(stats.episode_rewards) > 1 else \
                stats.episode_rewards[0]
        except IndexError:
            mean_eps_rewards = None
            mean_eps_length = None

        stdev_eps_length = statistics.stdev(stats.episode_lengths) if len(stats.episode_lengths) > 1 else 0
        stdev_eps_rewards = statistics.stdev(stats.episode_rewards) if len(stats.episode_rewards) > 1 else 0

        self.episode_length_history.append(stats.episode_lengths)
        self.episode_reward_history.append(stats.episode_rewards)
        self.cycle_length_history.append(None if stats.numb_completed_episodes == 0 else mean_eps_length)
        self.cycle_reward_history.append(None if stats.numb_completed_episodes == 0 else mean_eps_rewards)
        self.cycle_length_std_history.append(None if stats.numb_completed_episodes <= 1 else stdev_eps_length)
        self.cycle_reward_std_history.append(None if stats.numb_completed_episodes <= 1 else stdev_eps_rewards)
        self.cycle_stat_n_history.append(stats.numb_completed_episodes)
        self.underflow_history.append(stats.tbptt_underflow)
        self.current_per_receptor_mean = {s: arr.tolist() for s, arr in stats.per_receptor_mean.items()}

        for key, value in stats.auxiliary_performances.items():
            if key not in self.auxiliary_performances:
                self.auxiliary_performances[key] = {"mean": [], "std": []}
            self.auxiliary_performances[key]["mean"].append(np.mean(value).item())
            self.auxiliary_performances[key]["std"].append(np.std(value).item())

    def _make_actor(self, horizon, discount, lam, subseq_length) -> Gatherer:
        # create the Gatherer
        return self.gatherer_class(
            worker_id=self.comm_rank,
            exp_id=self.agent_id,
            distribution=self.distribution,
            horizon=horizon, discount=discount, lam=lam, subseq_length=subseq_length)

    def optimize(self, dataset: tf.data.TFRecordDataset, epochs: int, batch_size: int,
                 effective_batch_sizes: Union[int, Tuple[int, int]]) -> None:
        """Optimize the agent's policy and value network based on a given dataset.

        Args:
            dataset (tf.data.Dataset): tensorflow dataset containing serialization, a, p(a), r and A as components per data point
            epochs (int): number of epochs to train on this dataset
            batch_size (Tuple): batch size representing the number of TRANSITIONS per batch (not chunks)
            effective_batch_sizes (Tuple): batch size representing the number of TRAJECTORIES per batch and the
                                            number of CHUNKS per trajectory per batch

        Returns:
            None
        """

        # start optimization
        if self.is_recurrent:
            total_updates = self.n_updates_per_epoch * epochs
        else:
            total_updates = (self.n_workers * self.horizon) // batch_size * epochs
        policy_loss_history, value_loss_history, entropy_history = [], [], []
        with tqdm(total=total_updates, disable=not self.is_root, desc="Optimizing...") as pbar:
            for epoch in range(epochs):
                if self.is_recurrent:
                    batched_dataset = dataset.batch(effective_batch_sizes[0], drop_remainder=True)
                    with tf.device(self.device):
                        super_batch_ents = []
                        super_batch_pi_losses = []
                        super_batch_v_losses = []

                        for super_batch in batched_dataset:
                            ent, pi_loss, v_loss = recurrent_train_step(super_batch=super_batch,
                                                                        batch_size=effective_batch_sizes,
                                                                        joint=self.joint,
                                                                        distribution=self.distribution,
                                                                        continuous_control=self.continuous_control,
                                                                        clip_values=self.clip_values,
                                                                        gradient_clipping=self.gradient_clipping,
                                                                        clip=self.clip, c_value=self.c_value,
                                                                        c_entropy=self.c_entropy,
                                                                        is_recurrent=self.is_recurrent,
                                                                        optimizer=self.optimizer,
                                                                        pbar=pbar if self.is_root else None)

                            super_batch_ents.append(ent)
                            super_batch_pi_losses.append(pi_loss)
                            super_batch_v_losses.append(v_loss)

                        policy_loss_history.append(tf.reduce_mean(super_batch_pi_losses).numpy().item())
                        value_loss_history.append(tf.reduce_mean(super_batch_v_losses).numpy().item())
                        entropy_history.append(tf.reduce_mean(super_batch_ents).numpy().item())
                else:
                    dataset = dataset.shuffle(10000, reshuffle_each_iteration=True)

                    # divide into batches
                    batched_dataset = dataset.batch(batch_size, drop_remainder=True)
                    policy_epoch_losses, value_epoch_losses, entropies = [], [], []

                    with tf.device(self.device):
                        for b in batched_dataset:
                            ent, pi_loss, v_loss = ff_train_step(batch=b, joint=self.joint,
                                                                 distribution=self.distribution,
                                                                 continuous_control=self.continuous_control,
                                                                 clip_values=self.clip_values,
                                                                 gradient_clipping=self.gradient_clipping,
                                                                 clip=self.clip, c_value=self.c_value,
                                                                 c_entropy=self.c_entropy,
                                                                 is_recurrent=self.is_recurrent,
                                                                 optimizer=self.optimizer)
                            pbar.update(1)

                        entropies.append(ent)
                        policy_epoch_losses.append(pi_loss)
                        value_epoch_losses.append(v_loss)

                        # reset RNN states after each outer batch
                        self.joint.reset_states()

                    # remember some statistics
                    policy_loss_history.append(tf.reduce_mean(policy_epoch_losses).numpy().item())
                    value_loss_history.append(tf.reduce_mean(value_epoch_losses).numpy().item())
                    entropy_history.append(tf.reduce_mean(entropies).numpy().item())

        if MPI is not None:
            self.optimization_comm.Barrier()

        # store statistics in agent history
        self.policy_loss_history.append(statistics.mean(policy_loss_history))
        self.value_loss_history.append(statistics.mean(value_loss_history))
        self.entropy_history.append(statistics.mean(entropy_history))

        tf.keras.backend.clear_session()
        tf.compat.v1.reset_default_graph()
        gc.collect()

    def evaluate(self, n: int, save: bool = False, act_confidently=False) -> Tuple[StatBundle, Any]:
        """Evaluate the current state of the policy on the given environment for n episodes.

        Args:
            n (int): integer value indicating the number of episodes that shall be run
            save (bool): whether to save the evaluation to the monitor directory
            act_confidently (bool): Choose actions deterministically instead of sampling them from the distribution
        Returns:
            StatBundle with evaluation results
        """
        policy, value, joint = self.build_models(self.joint.get_weights(), 1, 1)

        stat_bundles = []
        for i in tqdm(range(n), disable=not self.is_root):
            lengths, rewards, classes, auxiliary_performances = evaluate(policy, self.env, self.distribution,
                                                                         act_confidently)
            stat_bundles.append(StatBundle(1, lengths, [rewards], [lengths], tbptt_underflow=0, per_receptor_mean={},
                                           auxiliary_performances=auxiliary_performances))

        stats = condense_stats(stat_bundles)

        if save and self.is_root:
            os.makedirs(f"{const.PATH_TO_EXPERIMENTS}/{self.agent_id}/", exist_ok=True)

            previous_evaluations = {}
            if os.path.isfile(f"{const.PATH_TO_EXPERIMENTS}/{self.agent_id}/evaluations.json"):
                with open(f"{const.PATH_TO_EXPERIMENTS}/{self.agent_id}/evaluations.json", "r") as jf:
                    previous_evaluations = json.load(jf)

            previous_evaluations.update({self.iteration: stats._asdict()})

            with open(f"{const.PATH_TO_EXPERIMENTS}/{self.agent_id}/evaluations.json", "w") as jf:
                json.dump(previous_evaluations, jf)

        return stats, classes

    def report(self, total_iterations, verbose=False):
        """Print a report of the current state of the training."""
        if not self.is_root:
            return

        sc, nc, ec, ac = COLORS["OKGREEN"], COLORS["OKBLUE"], COLORS["ENDC"], COLORS["FAIL"]
        reward_col = ac
        if hasattr(self.env.spec, "reward_threshold") and self.env.spec.reward_threshold is not None and \
                self.cycle_reward_history[0] is not None and self.cycle_reward_history[-1] is not None:
            half_way_there_threshold = (self.cycle_reward_history[0] + 0.5 * (
                    self.env.spec.reward_threshold - self.cycle_reward_history[0]))
            if self.env.spec.reward_threshold < self.cycle_reward_history[-1]:
                reward_col = COLORS["GREEN"]
            elif self.cycle_reward_history[-1] > half_way_there_threshold:
                reward_col = COLORS["ORANGE"]

        # calculate computation spend on different phases of the iteration and the percentages
        time_string = ""
        time_distribution_string = ""
        if len(self.time_dicts) > 0:
            times = [time for time in self.time_dicts[-1].values() if time is not None]
            time_percentages = [str(round(100 * t / sum(times))) for i, t in enumerate(times)]
            time_string = "[" + "|".join([str(round(t, 1)) for t in times]) + "]"
            time_distribution_string = "[" + "|".join(map(str, time_percentages)) + "]"
        if isinstance(self.lr_schedule, tf.keras.optimizers.schedules.LearningRateSchedule):
            current_lr = self.lr_schedule(self.optimizer.iterations)
        else:
            current_lr = self.lr_schedule

        # losses
        pi_loss = "  pi  " if len(self.policy_loss_history) == 0 else f"{round(self.policy_loss_history[-1], 2):6.2f}"
        v_loss = "  v     " if len(self.value_loss_history) == 0 else f"{round(self.value_loss_history[-1], 2):8.2f}"
        ent = "  ent " if len(self.entropy_history) == 0 else f"{round(self.entropy_history[-1], 2):6.2f}"

        # tbptt underflow
        underflow = f"w: {nc}{self.underflow_history[-1]}{ec}; " if self.underflow_history[-1] is not None else ""

        # timings
        time_left = "unknown time"
        if len(self.cycle_timings) > 0:
            time_left = f"{round(ignore_none(statistics.mean, self.cycle_timings) * (total_iterations - self.iteration) / 60, 1)}mins"

        total_gpu_memory = 0
        if len(self.gpus) > 0:
            nvidia_handle = nvidia_smi.nvmlDeviceGetHandleByIndex(0)
            nvidia_info = nvidia_smi.nvmlDeviceGetMemoryInfo(nvidia_handle)
            total_gpu_memory = nvidia_info.total

        # years of experience
        years_of_experience_report = ""
        if self.years_of_experience is not None:
            years_of_experience_report = f"y.exp: {nc}{round(self.years_of_experience, 3):3.3f}{ec}; "

        report_items = {
            "cycle": f"{sc}{f'Cycle {self.iteration:5d}/{total_iterations}' if self.iteration != 0 else 'Before Training'}{ec}",
            "reward": f"r: {reward_col}{'-' if self.cycle_reward_history[-1] is None else f'{round(self.cycle_reward_history[-1], 2):8.2f}'}{ec}",
            "length": f"len: {nc}{'-' if self.cycle_length_history[-1] is None else f'{round(self.cycle_length_history[-1], 2):8.2f}'}{ec}",
            "n": f"n: {nc}{'-' if self.cycle_stat_n_history[-1] is None else f'{self.cycle_stat_n_history[-1]:3d}'}{ec}",
            "loss": f"loss: [{nc}{pi_loss}{ec}|{nc}{v_loss}{ec}|{nc}{ent}{ec}]",
            "time": f"time: {time_string} {time_distribution_string}", "underflow": underflow,
            "lr": f"lr: {nc}{current_lr:.2e}{ec}",
            "updates": f"upd: {nc}{self.optimizer.iterations.numpy().item() - 1:6d}{ec}",
            "frames": f"f: {nc}{round(self.total_frames_seen / 1e3, 3):8.3f}{ec}k",
            "time_left": f"time left: {nc}{time_left}{ec}", "yoe": years_of_experience_report,
            "eps": f"eps: {nc}{self.total_episodes_seen:5d}{ec}",
            "took": f"took {self.cycle_timings[-1] if len(self.cycle_timings) > 0 else ''}s [{time_left} left]",
            "mem": f"mem: {self.used_memory[-1]}/{round(psutil.virtual_memory()[0] / 1e9)}|{self.used_gpu_memory[-1]}/{round(total_gpu_memory / 1e9, 2)}"}

        included_items = ["cycle", "reward", "length", "n", "loss", "updates", "yoe", "time", "time_left", "took"]
        if verbose:
            included_items = report_items.keys()

        mpi_flat_print("; ".join([report_items[k] for k in included_items]) + "\n")

    def save_agent_state(self, name=None):
        """Save the current state of the agent into the agent directory, identified by the current iteration."""
        if name is None:
            name = str(self.iteration)

        if not os.path.exists(self.agent_directory + f"/{name}/weights"):
            os.makedirs(self.agent_directory + f"/{name}/weights")

        self.joint.save_weights(os.path.join(self.agent_directory, f"{name}/weights"), overwrite=True)
        with open(self.agent_directory + f"/{name}/parameters.json", "w") as f:
            json.dump(self.get_parameters(), f)

        np.savez(self.agent_directory + f"/{name}/optimizer_weights.npz", *self.optimizer.get_weights())

    def get_parameters(self):
        """Get the agents parameters necessary to reconstruct it."""
        parameters = self.__dict__.copy()
        del parameters["env"]
        del parameters["policy"], parameters["value"], parameters["joint"], parameters["distribution"]
        del parameters["optimizer"], parameters["lr_schedule"], parameters["model_builder"], parameters[
            "gatherer_class"]
        del parameters["mpi_comm"], parameters["optimization_comm"], parameters["is_optimization_process"], parameters[
            "n_optimizers"], parameters["gpus"], parameters["is_root"]

        parameters["c_entropy"] = parameters["c_entropy"].numpy().item()
        parameters["c_value"] = parameters["c_value"].numpy().item()
        parameters["clip"] = parameters["clip"].numpy().item()
        parameters["distribution"] = self.distribution.__class__.__name__
        parameters["transformers"] = self.env.serialize()
        parameters["optimizer"] = self.optimizer.serialize()

        return parameters

    @staticmethod
    def from_agent_state(agent_id: int, from_iteration: Union[int, str] = None, force_env_name=None, path_modifier="",
                         n_optimizers: int = None) -> "PPOAgent":
        """Build an agent from a previously saved state.

        Args:
            agent_id:           the ID of the agent to be loaded
            from_iteration:     from which iteration to load, if None (default) use most recent, can be iteration int
                                or ["b", "best"] for best, if such was saved.

        Returns:
            loaded_agent: a PPOAgent object of the same state as the one saved into the path specified by agent_id
        """
        if MPI is not None:
            mpi_comm = MPI.COMM_WORLD
            is_root = mpi_comm.rank == 0
            optimization_comm, is_optimization_process = PPOAgent.get_optimization_comm(
                limit_to_n_optimizers=n_optimizers)
        else:
            is_root = True
            optimization_comm, is_optimization_process = None, True

        # TODO also load the state of the optimizers
        agent_path = path_modifier + BASE_SAVE_PATH + f"/{agent_id}" + "/"

        if not os.path.isdir(agent_path):
            raise FileNotFoundError(
                f"The given agent ID does not match any existing save history from your current path. Searched for "
                f"{os.path.abspath(agent_path)}")

        if len(os.listdir(agent_path)) == 0:
            raise FileNotFoundError("The given agent ID'serialization save history is empty.")

        # determine loading point
        latest_matches = PPOAgent.get_saved_iterations(agent_id, path_modifier=path_modifier)

        if from_iteration is None:
            if len(latest_matches) > 0:
                from_iteration = max(latest_matches)
            else:
                from_iteration = "best"
        elif isinstance(from_iteration, str):
            assert from_iteration.lower() in ["best", "b",
                                              "last"], "Unknown string identifier, can only be 'best'/'b'/'last' or int."
            if from_iteration == "b":
                from_iteration = "best"
            if from_iteration == "last" and not os.path.isdir(f"{agent_path}/last"):
                from_iteration = "best"
        else:
            assert from_iteration in latest_matches, "There is no save at this iteration."

        # make stack of fallbacks in case the targeted iteration is corrupted
        fallback_stack = ["best", "last"]
        if len(latest_matches) > 0:
            fallback_stack.append(max(latest_matches))

        can_load = False
        while not can_load:
            try:
                loading_path = f"{agent_path}/{from_iteration}/"
                with open(f"{loading_path}/parameters.json", "r") as f:
                    parameters = json.load(f)

                can_load = True
            except json.decoder.JSONDecodeError as e:
                print(f"The parameter file of {from_iteration} seems to be corrupted. "
                      f"Falling back to {fallback_stack[0]}.")
                print(e)
                from_iteration = fallback_stack.pop(0)

        if is_root:
            print(f"Loading from iteration {from_iteration}.")

        env = make_task(parameters["env_name"] if force_env_name is None else force_env_name,
                        reward_config=parameters.get("reward_configuration"),
                        transformers=postprocessors_from_serializations(parameters["transformers"]),
                        render_mode="rgb_array" if re.match(".*[Vv]is(ion|ual).*", parameters["env_name"]) else None)
        model_builder = models.MODEL_BUILDERS[parameters["builder_function_name"]]
        distribution = getattr(policies, parameters["distribution"])(env)

        loaded_agent = PPOAgent(model_builder, environment=env, horizon=parameters["horizon"],
                                workers=parameters["n_workers"], learning_rate=parameters["learning_rate"],
                                discount=parameters["discount"], lam=parameters["lam"], clip=parameters["clip"],
                                c_entropy=parameters["c_entropy"], c_value=parameters["c_value"],
                                gradient_clipping=parameters["gradient_clipping"],
                                clip_values=parameters["clip_values"], tbptt_length=parameters["tbptt_length"],
                                lr_schedule=parameters["lr_schedule_type"], distribution=distribution, _make_dirs=False,
                                reward_configuration=parameters["reward_configuration"], n_optimizers=n_optimizers)

        for p, v in parameters.items():
            if p in ["distribution", "transformers", "c_entropy", "c_value", "gradient_clipping", "clip", "optimizer"]:
                continue

            loaded_agent.__dict__[p] = v

        loaded_agent.joint.load_weights(
            f"./{path_modifier}/{BASE_SAVE_PATH}/{agent_id}/" + f"/{from_iteration}/weights")

        if "optimizer" in parameters.keys():  # for backwards compatibility
            if os.path.isfile(agent_path + f"/{from_iteration}/optimizer_weights.npz"):
                optimizer_weights = list(np.load(agent_path + f"/{from_iteration}/optimizer_weights.npz").values())
                parameters["optimizer"]["weights"] = optimizer_weights

                loaded_agent.optimizer = MpiAdam.from_serialization(optimization_comm, parameters["optimizer"],
                                                                    loaded_agent.joint.trainable_variables)
            else:
                loaded_agent.optimizer = MpiAdam.from_serialization(optimization_comm, parameters["optimizer"],
                                                                    loaded_agent.joint.trainable_variables)

        # mark the loading
        loaded_agent.loading_history.append([loaded_agent.iteration])

        return loaded_agent

    @staticmethod
    def get_saved_iterations(agent_id: int, path_modifier="") -> list:
        """Return a list of iterations at which the agent of given ID has been saved."""
        agent_path = os.path.join(path_modifier, BASE_SAVE_PATH, f"{agent_id}")

        if not os.path.isdir(agent_path):
            raise FileNotFoundError("The given agent ID does not match any existing save history.")

        if len(os.listdir(agent_path)) == 0:
            return []

        its = [int(i.group(0)) for i in [re.match("([0-9]+)", fn) for fn in os.listdir(agent_path)] if i is not None]

        return sorted(its)

    def finalize(self):
        """Perform final steps on the agent that are necessary no matter whether an error occurred or not."""
        for file in glob(fr"{STORAGE_DIR}/{self.agent_id}_data_[0-9]+\.tfrecord"):
            os.remove(file)

    def build_models(self, weights, batch_size, sequence_length=1):
        """Build the models (policy, value, joint) with the provided weights."""
        policy, value, joint = self.model_builder(self.env, self.distribution,
                                                  **({"bs": batch_size} if requires_batch_size(
                                                      self.model_builder) else {}),
                                                  **({"sequence_length": sequence_length} if requires_sequence_length(
                                                      self.model_builder) else {}))
        joint.set_weights(weights)

        return policy, value, joint

    def load_components(self, path_to_components: str):
        for directory in os.listdir(path_to_components):
            pretrained_component_full_path = os.path.join(path_to_components, directory)
            if os.path.isdir(pretrained_component_full_path):
                try:
                    pretrained_component = tf.keras.models.load_model(pretrained_component_full_path, compile=False)
                except JSONDecodeError as e:
                    raise ComponentError("Could not load the pretrained model from given files. Likely, the saved model"
                                         "files where created with a different incompatible TF version.")

                found_counterpart = False
                for model_component in self.joint.layers:
                    if model_component.name == pretrained_component.name:
                        mpi_print(f"Loading weights of '{model_component.name}' from pretrained models.")
                        model_component.set_weights(pretrained_component.get_weights())

                        found_counterpart = True
                        break

                if not found_counterpart:
                    print(f"No counterpart for pretrained component '{pretrained_component.name}' found in the model.")

    def freeze_component(self, component: str):
        for model_component in self.joint.layers:
            if model_component.name == component:
                mpi_print(f"Freezing '{model_component.name}'.")
                model_component.trainable = False



import angorapy as ap
import angorapy.tasks.core
import angorapy.tasks.registration
from angorapy.tasks.wrappers import TaskWrapper


def _test_any_task(task: TaskWrapper):
    state = task.reset()
    for _ in range(100):
        state, r, dterm, dtrunc, info = task.step(task.action_space.sample())

        if dtrunc or dterm:
            state = task.reset()


def test_manipulate():
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlock-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockVisual-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockAsymmetric-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockVisualAsymmetric-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockDiscrete-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockVisualDiscrete-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockDiscreteAsymmetric-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ManipulateBlockVisualDiscreteAsymmetric-v0"))


def test_reach():
    _test_any_task(angorapy.tasks.registration.make_task("ReachAbsolute-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ReachRelative-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ReachAbsoluteVisual-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("ReachRelativeVisual-v0"))


def test_free_reach():
    _test_any_task(angorapy.tasks.registration.make_task("FreeReachAbsolute-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("FreeReachRelative-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("FreeReachAbsoluteVisual-v0"))
    _test_any_task(angorapy.tasks.registration.make_task("FreeReachRelativeVisual-v0"))


def test_cognitive():
    _test_any_task(angorapy.tasks.registration.make_task("HanoiTower-v0"))


from typing import Sequence

from dm_control import mjcf
from mujoco_utils import mjcf_utils

from angorapy.tasks.world_building.entities import Robot
from angorapy.tasks.envs.dexterity.mujoco_model import consts

HAND_POSITION = (0., 0., 0.)
HAND_QUATERNION = (1, -1, 1, -1)


class ShadowHand(Robot):
    """A ShadowHand robot implementation."""

    def __init__(self,
                 position: Sequence[float] = HAND_POSITION,
                 rotation: Sequence[float] = HAND_QUATERNION,
                 no_wrist_control: bool = False) -> None:
        """Initializes a ShadowHand.

        Args:
            position: The position of the robot. Defaults to (0, 0, 0).
            rotation: The rotation of the robot, as a quaternion. Defaults to (1, -1, 1, -1), palm facing up.
            no_wrist_control: If True, the wrist is fixed and cannot be controlled. Defaults to False.
        """
        self._prefix = "rh_"
        self.no_wrist_control = no_wrist_control
        super().__init__(consts.MODEL_XML, position=position, rotation=rotation)

    def _parse_entity(self) -> None:
        if self.no_wrist_control:
            self._delete_wrist_actuators()

        joints = mjcf_utils.safe_find_all(self._mjcf_root, "joint")
        actuators = mjcf_utils.safe_find_all(self._mjcf_root, "actuator")

        self._joints = tuple(joints)
        self._actuators = tuple(actuators)

    def _setup_entity(self):
        pass

    def _delete_wrist_actuators(self):
        actuators = mjcf_utils.safe_find_all(self._mjcf_root, "actuator")
        for actor in actuators:
            if "_WR" in actor.name:
                actor.remove()

    # VISUALIZATION
    def show_touch_sites(self, show: bool = True):
        for site in self.mjcf_model.worldbody.find_all("site"):
            if "T_" in site.name:
                site.rgba[-1] = 1 if show else 0

    def show_palm_site(self, show: bool = True):
        for site in self.mjcf_model.worldbody.find_all("site"):
            if "palm_center_site" in site.name:
                site.rgba[-1] = 1 if show else 0

    # ACCESSORS
    @property
    def root_body(self) -> mjcf.Element:
        return mjcf_utils.safe_find(self._mjcf_root, "body", self._prefix + "forearm")

    @property
    def joints(self) -> Sequence[mjcf.Element]:
        return self._joints

    @property
    def actuators(self) -> Sequence[mjcf.Element]:
        return self._actuators


class ShadowHandReach(ShadowHand):
    """A Shadow Hand E3M5."""

    def _setup_entity(self):
        super()._setup_entity()

        colors = [[1, 0, 0, 1], [0, 1, 0, 1], [0, 0, 1, 1], [1, 1, 0, 1], [1, 0, 1, 1]]

        indicator_body = self.mjcf_model.worldbody
        for finger_idx in range(5):
            # note that initial positions need to be non-[0 0 0] or otherwise they cannot for some reason be changed
            # in simulation. wasted 5h of my life to find this out.

            indicator_body.add(
                "site",
                name=f"target{finger_idx}",
                pos=[0.01, 0, 0],
                rgba=colors[finger_idx],
                size=[0.005],
                type="sphere"
            )

            indicator_body.add(
                "site",
                name=f"finger{finger_idx}",
                pos=[0.01, 0, 0],
                rgba=colors[finger_idx],
                size=[0.01],
                type="sphere"
            )


if __name__ == '__main__':
    shadow_hand = ShadowHand(no_wrist_control=True)



"""Wrappers encapsulating envs to modulate n_steps, rewards, and control state initialization."""
import abc
from typing import List, OrderedDict, Tuple, Dict, Any, SupportsFloat

import gymnasium as gym
import numpy

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

from angorapy.common.senses import Sensation
from angorapy.common.postprocessors import BasePostProcessor, merge_postprocessors


class TaskWrapper(gym.ObservationWrapper):
    """Wrapper for all tasks with basic functionality.

    Parses observations of the wrapped task to Sensations. Serves additionally as base class for all other wrappers.

    Args:
        env (gym.Env): The task to wrap.

    Attributes:
        env (gym.Env): The wrapped environment.
    """

    def __init__(self, env: gym.Env):
        super().__init__(env)

    @property
    def name(self):
        """The name of the transformers."""
        return self.__class__.__name__

    @abc.abstractmethod
    def warmup(self, n_steps=10):
        """Warmup the environment."""
        pass

    def reset(self, **kwargs):
        """Resets the environment, returning a modified observation using :meth:`self.observation`."""
        obs, info = self.env.reset(**kwargs)
        return self.observation(obs), self.info(info)

    def observation(self, observation) -> Sensation:
        """Process an observation to be of type 'Sensation'."""
        if isinstance(observation, Sensation):
            return observation
        elif isinstance(observation, (dict, OrderedDict)):
            assert "observation" in observation.keys(), "Unknown dict type of state couldnt be resolved to Sensation."

            if isinstance(observation["observation"], Sensation):
                return observation["observation"]
            elif isinstance(observation["observation"], numpy.ndarray):  # GOAl ENVS
                return Sensation(proprioception=observation["observation"], goal=observation["desired_goal"])
            elif isinstance(observation["observation"], dict) and all(
                    [k in observation["observation"].keys() for k in Sensation.sense_names]):
                return Sensation(**observation["observation"])

        return Sensation(proprioception=observation)

    def step(self, action) -> Tuple[Sensation, SupportsFloat, bool, bool, Dict[str, Any]]:
        observation, reward, terminated, truncated, info = self.env.step(action)

        if hasattr(observation, "keys") and "achieved_goal" in observation.keys():
            info["achieved_goal"] = observation["achieved_goal"]

        if hasattr(observation, "keys") and "desired_goal" in observation.keys():
            info["desired_goal"] = observation["desired_goal"]

        return self.observation(observation), reward, terminated, truncated, info

    def info(self, info):
        return info

    # SYNCHRONIZATION

    def mpi_sync(self):
        """Synchronize the wrapper and all wrapped wrappers over all MPI ranks."""
        if isinstance(self.env, TaskWrapper):
            self.env.mpi_sync()

        self._mpi_sync()

    @abc.abstractmethod
    def _mpi_sync(self):
        pass

    # SERIALIZATION

    @abc.abstractmethod
    def serialize(self):
        """Serialize the wrappers defining data."""
        pass


class TransformationWrapper(TaskWrapper):
    """Wrapper transforming rewards and observation based on running means."""

    def __init__(self, env, transformers: List[BasePostProcessor]):
        super().__init__(env)

        # TODO maybe change this to expect list of types to build itself?
        self.transformers = transformers

    def __contains__(self, item):
        return item in self.transformers

    def step(self, action):
        """Perform a step and transform the results."""
        step_tuple = super().step(action)

        # include original reward in info
        step_tuple[-1]["original_reward"] = step_tuple[1]

        if len(self.transformers) != 0:
            for transformer in self.transformers:
                step_tuple = transformer.transform(step_tuple, update=True)

        return step_tuple

    def add_transformers(self, transformers):
        """Add a list of transformers to the environment."""
        self.transformers.extend(transformers)

    def clear_transformers(self):
        """Clear the list of transformers."""
        self.transformers = []

    def warmup(self, n_steps=10):
        """Warmup the transformers."""
        for t in self.transformers:
            t.warmup(self, n_steps=n_steps)

    def _mpi_sync(self):
        """Synchronise the transformers of the wrapper over all MPI ranks."""
        if MPI is None:
            return

        synced_transformers = []
        for transformer in self.transformers:
            collection = MPI.COMM_WORLD.gather(transformer, root=0)

            if MPI.COMM_WORLD.Get_rank() == 0:
                synced_transformers.append(merge_postprocessors(collection))

        self.transformers = MPI.COMM_WORLD.bcast(synced_transformers, root=0)

    def serialize(self):
        """Return separate transformer serializations in a list"""
        return [t.serialize() for t in self.transformers]



import abc
from abc import ABC
from os import path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

import dm_control
import gymnasium as gym
import mujoco
import numpy as np
from dm_control.mjcf import RootElement
from gymnasium import spaces
from gymnasium.utils import seeding
from mujoco import MjModel

from angorapy.common.const import N_SUBSTEPS
from angorapy.common.const import VISION_WH
from angorapy.common.senses import Sensation
from angorapy.tasks.reward_config import resolve_config_name
from angorapy.tasks.utils import convert_observation_to_space
from angorapy.tasks.utils import mj_get_category_names
from angorapy.tasks.utils import mj_qpos_dict_to_qpos_vector
from angorapy.tasks.world_building.entities import _Entity


def _parse_model_definition(model_definition: Union[str, RootElement, _Entity]) -> Tuple[MjModel, RootElement]:
    """Parse a variable model definition into a MuJoCo model and a MJCF Model.

    Possible model definitions are:
        - a path to a model xml file
        - a dm_control.mjcf.RootElement object
        - a angorapy.tasks.world_building.entities._Entity object

    Args:
        model_definition (Union[str, RootElement, _Entity]): The model definition to parse.

    Returns:
        Tuple[MjModel, RootElement]: The parsed MuJoCo model and MuJoCo XML model.
    """
    if isinstance(model_definition, str):
        model_path = model_definition
        if model_path.startswith("/"):
            fullpath = model_path
        else:
            fullpath = path.join(path.dirname(__file__), "models", model_path)

        if not path.exists(fullpath):
            raise OSError(f"File {fullpath} does not exist")

        model = mujoco.MjModel.from_xml_path(filename=fullpath)
        mjcf_model = dm_control.mjcf.from_path(fullpath)
    elif isinstance(model_definition, dm_control.mjcf.RootElement):
        model = mujoco.MjModel.from_xml_string(xml=model_definition.to_xml_string(),
                                               assets=model_definition.get_assets())
        mjcf_model = model_definition
    elif isinstance(model_definition, _Entity):
        model = mujoco.MjModel.from_xml_string(xml=model_definition.mjcf_model.to_xml_string(),
                                               assets=model_definition.mjcf_model.get_assets())
        mjcf_model = model_definition._mjcf_root.root_model
    else:
        raise ValueError(f"model must be either a path to a model xml file or mjcf.RootElement, "
                         f"not {type(model_definition)}")

    return model, mjcf_model


class AnthropomorphicEnv(gym.Env, ABC):
    """Base class for all anthropomorphic environments.

    Connects to a MuJoCo model and its simulation and provides basic functionality for an anthropomorphic environment.

    Args:
        model_path (str): Path to the model xml file.
        frame_skip (int): Number of frames to skip per action.
        initial_qpos (dict): Dictionary of initial joint positions.
        vision (bool): Whether to use vision or not.
        touch (bool): Whether to use touch or not.
        render_mode (str): The render mode to use, in ["human", "rgb_array"].
        camera_id (int): The camera id to use.
        camera_name (str): The camera name to use.
        delta_t (float): The time between two control steps.
        n_substeps (int): The number of simulation steps per control step.

    Attributes:
        metadata (dict): The metadata of the environment.
        observation_space (gym.spaces.Space): The observation space of the environment.
        action_space (gym.spaces.Space): The action space of the environment.
        reward_range (tuple): The reward range of the environment.
        spec (gym.envs.registration.EnvSpec): The specification of the environment.
        viewer (mujoco_py.MjViewer): The viewer of the environment, used for rendering in different modes.
        model (mujoco_py.MjModel): The MuJoCo model used for simulation.
        data (mujoco_py.MjData): The MuJoCo data object containing simulation data.
        initial_state (dict): The initial state of the environment.
        goal (np.ndarray): The goal of the environment.
        vision (bool): Whether to use vision or not.
        touch (bool): Whether to use touch or not.
        render_mode (str): The render mode to use, in ["human", "rgb_array"].
        camera_id (int): The camera id to use.
        camera_name (str): The camera name to use.
        frame_skip (int): Number of frames to skip per action.
    """
    metadata: Dict[str, Any] = {
        "render_modes": ["human", "rgb_array"]}

    continuous = True
    discrete_bin_count = 11

    def __init__(
            self,
            model: Union[str, RootElement, _Entity],
            initial_qpos=None,
            vision=False,
            touch=True,
            relative_control=True,
            render_mode: Optional[str] = None,
            camera_id: Optional[int] = 0,
            camera_name: Optional[str] = None,
            delta_t: float = 0.002,
            frame_skip=N_SUBSTEPS,
            n_substeps: int = N_SUBSTEPS,
    ):
        self.model, self.mjcf_model = _parse_model_definition(model)

        self.model.vis.global_.offwidth = VISION_WH
        self.model.vis.global_.offheight = VISION_WH
        self.data = mujoco.MjData(self.model)

        # time control
        self._delta_t_control: float = delta_t
        self._delta_t_simulation: float = delta_t
        self._simulation_steps_per_control_step: int = int(self._delta_t_control // self._delta_t_simulation)

        # rendering
        self.render_mode = render_mode
        self.camera_id = camera_id
        self.camera_name = camera_name
        self.frame_skip = frame_skip

        self._viewers = {}
        self.viewer = None

        self.metadata = {
            "render_modes": ["human", "rgb_array", "depth_array"],
            "render_fps": int(np.round(1.0 / self.dt)),
        }

        # senses
        self.vision = vision
        self.touch = touch

        self._touch_sensor_id_site_id = []
        self._touch_sensor_id = []

        # store initial state
        initial_qpos = self.set_initial_qpos(initial_qpos)
        self.initial_state = {
            "qpos": initial_qpos,
            "qvel": self.data.qvel[:]
        }

        # setup environment
        self.relative_control = relative_control
        self._env_setup(initial_state=self.initial_state)
        self._set_action_space()

        self.goal = self._sample_goal()
        observation = self._get_obs()
        self._set_observation_space(observation)

        self.model.opt.timestep = delta_t
        self.original_n_substeps = n_substeps

        self._set_default_reward_function_and_config()
        self.seed()
        self.viewer_setup()

    def seed(self, seed=None):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    # SPACES
    def _set_action_space(self):
        """Sets the action space of the environment.

        Action spaces may be continuous or discrete. If continuous, the action space is a Box space with the bounds
        given by the model's actuator control range. If discrete, the action space is a MultiDiscrete space splitting
        the continuous action space into a number of bins (default 11)."""

        bounds = self.model.actuator_ctrlrange.copy().astype(np.float32)
        low, high = bounds.T

        if self.continuous:
            self.action_space = spaces.Box(-1., 1., shape=(self.model.actuator_ctrlrange.shape[0],), dtype=float)
        else:
            self.action_space = spaces.MultiDiscrete(
                np.ones(self.model.actuator_ctrlrange.shape[0]) * self.discrete_bin_count
            )
            self.discrete_action_values = np.linspace(-1, 1, self.discrete_bin_count)

        return self.action_space

    def _set_observation_space(self,
                               observation):
        self.observation_space = convert_observation_to_space(observation)
        return self.observation_space

    # REWARD
    @abc.abstractmethod
    def _set_default_reward_function_and_config(self):
        """Sets the default reward function and config.

        The reward config is optional (can be empty) but can be used to
        configure the reward function. The reward function is a function that takes the environment and the steps info
        dictionary and returns a reward value."""
        pass

    def set_reward_function(self,
                            function: Union[str, Callable]):
        """Set the environment reward function by its config identifier or a callable."""
        if isinstance(function, str):
            try:
                function = getattr(reward, function.split(".")[0])
            except AttributeError:
                raise AttributeError("Reward function unknown.")
        elif isinstance(function, Callable):
            pass
        else:
            raise ValueError("Unknown format for given reward function. Provide a string or a Callable.")

        self.reward_function = function

    def set_reward_config(self,
                          new_config: Union[str, dict]):
        """Set the environment's reward configuration by its identifier or a dict."""
        if isinstance(new_config, str):
            new_config: dict = resolve_config_name(new_config)

        self.reward_config = new_config
        if "SUCCESS_DISTANCE" in self.reward_config.keys():
            self.distance_threshold = self.reward_config["SUCCESS_DISTANCE"]

        self.assert_reward_setup()

    @abc.abstractmethod
    def assert_reward_setup(self):
        pass

    def compute_reward(self, info):
        """Compute reward with additional success bonus."""
        return self.reward_function(self, info)

    def _is_success(self, achieved_goal, desired_goal):
        return False

    # TIME
    @property
    def dt(self):
        return self.model.opt.timestep * self.frame_skip

    def set_delta_t_simulation(self,
                               new: float):
        """Set new value for the simulation delta t."""
        assert np.isclose(self._delta_t_control % new, 0, rtol=1.e-3, atol=1.e-4), \
            f"Delta t of simulation must divide control delta t into integer " \
            f"parts, but gives {self._delta_t_control % new}."

        self._delta_t_simulation = new
        self.model.opt.timestep = self._delta_t_simulation

        self._simulation_steps_per_control_step = int(self._delta_t_control / self._delta_t_simulation *
                                                      self.original_n_substeps)

    def set_original_n_substeps_to_sspcs(self):
        self.original_n_substeps = self._simulation_steps_per_control_step

    # SETUP
    @abc.abstractmethod
    def _env_setup(self, initial_state):
        raise NotImplementedError

    def viewer_setup(self):
        """
        This method is called when the viewer is initialized.
        Optionally implement this method, if you need to tinker with camera position and so forth.
        """

    def reset(
            self,
            *,
            seed: Optional[int] = None,
            return_info: bool = False,
            options: Optional[dict] = None,
    ):
        super().reset(seed=seed)

        did_reset_sim = False
        while not did_reset_sim:
            did_reset_sim = self._reset_sim()

        self.goal = self._sample_goal()
        obs = self._get_obs()
        info = self._get_info()

        if self.render_mode == "human":
            self.render()

        return obs, info

    # CONTROL
    def step(self, action: np.ndarray):
        if self.continuous:
            action = np.clip(action, self.action_space.low, self.action_space.high)
        else:
            action = self.discrete_action_values[action.astype(int)]

        # perform simulation
        self.do_simulation(action, n_frames=self.original_n_substeps)

        # read out observation from simulation
        obs = self._get_obs()

        done = False
        info = self._get_info()

        reward = self.compute_reward(info)

        if self.render_mode == "human":
            self._render_callback()
            self.render()

        return obs, reward, done, False, info

    def _set_action(self, action):
        if np.array(action).shape != self.action_space.shape:
            raise ValueError(f"Action dimension mismatch [{np.array(action).shape}|{self.action_space.shape}]!")

        actuator_names, actuator_adresses = mj_get_category_names(self.model, "actuator")
        ctrlrange = self.model.actuator_ctrlrange
        actuation_range = (ctrlrange[:, 1] - ctrlrange[:, 0]) / 2.
        if self.relative_control:
            actuation_center = np.zeros_like(action)
            for i in range(self.data.ctrl.shape[0]):
                jnt_or_ten_name = actuator_names[i].replace(b'_A_', b'_')
                if self.model.names.index(jnt_or_ten_name) in self.model.name_jntadr:
                    actuation_center[i] = self.data.jnt(jnt_or_ten_name).qpos
                elif self.model.names.index(jnt_or_ten_name) in self.model.name_tendonadr:
                    actuation_center[i] = self.data.actuator(actuator_names[i]).length
        else:
            actuation_center = (ctrlrange[:, 1] + ctrlrange[:, 0]) / 2.

        self.data.ctrl[:] = actuation_center + action * actuation_range
        self.data.ctrl[:] = np.clip(self.data.ctrl,
                                    ctrlrange[:, 0],
                                    ctrlrange[:, 1])

    def do_simulation(self, ctrl, n_frames):
        self._set_action(ctrl)

        for _ in range(n_frames):
            mujoco.mj_step(self.model, self.data)
            self._mj_step_callback()

        mujoco.mj_rnePostConstraint(self.model, self.data)

    def _mj_step_callback(self):
        """A custom callback that is called after stepping the simulation. Can be used
        to enforce additional constraints on the simulation state.
        """
        pass

    # RENDERING
    def _get_viewer(
            self,
            mode
    ) -> Union[
        "gym.envs.mujoco.mujoco_rendering.Viewer",
        "gym.envs.mujoco.mujoco_rendering.RenderContextOffscreen",
    ]:
        self.viewer = self._viewers.get(mode)
        if self.viewer is None:
            if mode == "human":
                from gymnasium.envs.mujoco.mujoco_rendering import WindowViewer

                self.viewer = WindowViewer(self.model, self.data)
            elif mode in {"rgb_array", "depth_array"}:
                from gymnasium.envs.mujoco.mujoco_rendering import OffScreenViewer

                self.viewer = OffScreenViewer(self.model, self.data)
            else:
                raise AttributeError(
                    f"Unexpected mode: {mode}, expected modes: {self.metadata['render_modes']}"
                )

            self.viewer_setup()
            self._viewers[mode] = self.viewer

        return self.viewer

    def render(self):
        if self.render_mode is None:
            return

        self._render_callback()

        if self.render_mode in {
            "rgb_array",
            "depth_array",
        }:
            # camera_id = self.camera_id
            # camera_name = self.camera_name
            #
            # if camera_id is not None and camera_name is not None:
            #     raise ValueError(
            #         "Both `camera_id` and `camera_name` cannot be"
            #         " specified at the same time."
            #     )
            #
            # no_camera_specified = camera_name is None and camera_id is None
            # if no_camera_specified:
            #     camera_name = "track"
            #
            # if camera_id is None:
            #     camera_id = mujoco.mj_name2id(
            #         self.model,
            #         mujoco.mjtObj.mjOBJ_CAMERA,
            #         camera_name,
            #     )

            self._get_viewer(self.render_mode).render(render_mode=self.render_mode)

        if self.render_mode == "rgb_array":
            data = self._get_viewer(self.render_mode).render(camera_id=-1, render_mode=self.render_mode)
            # original image is upside-down, so flip it
            # return data[::-1, :, :]
            return data
        elif self.render_mode == "depth_array":
            self._get_viewer(self.render_mode).render()
            # Extract depth part of the render() tuple
            data = self._get_viewer(self.render_mode).render(depth=True)[1]
            # original image is upside-down, so flip it
            return data[::-1, :]
        elif self.render_mode == "human":
            self._get_viewer(self.render_mode).render()

    def close(self):
        if self.viewer is not None:
            self.viewer = None
            self._viewers = {}

    # SIMULATION STATE
    def state_vector(self):
        return np.concatenate([self.data.qpos.flat, self.data.qvel.flat])

    def get_state(self):
        """Get the current state of the simulation."""
        return {
            "qpos": self.data.qpos[:],
            "qvel": self.data.qvel[:],
        }

    def get_robot_state(self):
        """Returns all joint positions and velocities associated with a robot."""
        joint_names, _ = mj_get_category_names(self.model, "jnt")
        if self.data.qpos is not None and joint_names:
            names = [n for n in joint_names if n.startswith(b"robot")]
            return (
                np.array([self.data.jnt(name).qpos for name in names]).flatten(),
                np.array([self.data.jnt(name).qvel for name in names]).flatten(),
            )
        return np.zeros(0), np.zeros(0)

    def set_state(self, qpos, qvel):
        assert qpos.shape == (self.model.nq,) and qvel.shape == (self.model.nv,), \
            f"State shape [{qpos.shape}|{qvel.shape}] does not fit [{self.model.nq}|{self.model.nv}]"

        self.data.qpos[:] = np.copy(qpos)
        self.data.qvel[:] = np.copy(qvel)
        if self.model.na == 0:
            self.data.act[:] = None
        mujoco.mj_forward(self.model, self.data)

    def _reset_sim(self):
        """Reset the simulation."""
        self.reset_model()
        return True

    def reset_model(self):
        mujoco.mj_resetData(self.model, self.data)
        self.set_state(**self.initial_state)

        return True

    # SENSES
    def get_proprioception(self):
        """Get proprioception sensor readings."""
        robot_qpos, robot_qvel = self.get_robot_state()
        proprioception = np.concatenate([robot_qpos, robot_qvel])

        return proprioception

    def get_touch(self):
        """Get touch sensor readings."""
        touch = self.data.sensordata[self._touch_sensor_id]

        return touch

    def get_vision(self):
        """Get vision sensor readings."""
        if self.vision:
            tmp_render_mode = self.render_mode
            self.render_mode = "rgb_array"
            vision_input = self.render()
            self.render_mode = tmp_render_mode
        else:
            vision_input = np.array([])

        return vision_input

    def _get_obs(self):
        """Get proprioception, touch and vision sensor readings."""
        return {
            'observation': Sensation(
                proprioception=self.get_proprioception(),
                touch=self.get_touch() if self.touch else None,
                vision=self.get_vision() if self.vision else None,
                goal=self.goal.copy()
            ),
        }

    def _get_info(self):
        """Get dict containing additional information about the current state of the environment."""
        return {
            "auxiliary_performances": {}
        }

    # ABSTRACT METHODS
    @abc.abstractmethod
    def _sample_goal(self) -> np.ndarray:
        raise NotImplementedError("Must be implemented in subclass.")

    def _render_callback(self, **kwargs):
        """A callback that is called before rendering. Can be used to implement custom rendering."""
        pass

    def set_initial_qpos(self, initial_qpos):
        """Set the initial qpos of the simulation based on given configuration. If the configuration is only specific
        to the robot, the rest of the qpos is left unchanged."""

        if initial_qpos is not None and len(initial_qpos) == 24:
            final_initial_qpos = self.data.qpos.copy()
            final_initial_qpos[0:24] = mj_qpos_dict_to_qpos_vector(self.model, initial_qpos)

            return final_initial_qpos
        elif initial_qpos is None or not initial_qpos:
            initial_qpos = self.data.qpos[:]

        return initial_qpos

    # MODEL PROPERTIES
    @property
    def actuator_names(self):
        return mj_get_category_names(self.model, "actuator")[0]

    @property
    def joint_names(self):
        return mj_get_category_names(self.model, "jnt")[0]



"""Predefined configurations for different (groups of) envs."""


def make_config(batch_size=128, c_entropy=0.01, c_value=1, clip=0.2, cpu=False, debug=False, discount=0.99,
                env='CartPole-v1', epochs=3, eval=False, export_file=None, grad_norm=0.5, horizon=2024,
                iterations=100, lam=0.97, load_from=None, lr_pi=0.001, clip_values=False, save_every=0,
                workers=8, tbptt: int = 16, lr_schedule=None, no_state_norming=False, no_reward_norming=False,
                model="ffn", early_stopping=False, distribution=None, shared=False, preload=None, sequential=False,
                architecture="simple", radical_evaluation=False, redis_ip=None):
    """Make a config from scratch."""
    return dict(**locals())


def derive_config(original: dict, overrules: dict):
    """Make a new config with another config as base, overruling a given set of parameters."""
    derived = original.copy()
    for k, v in overrules.items():
        derived[k] = v
    return derived


test = make_config(
    batch_size=64,
    horizon=64,
    c_entropy=0.01,
    lr_pi=0.0003,
    epochs=2,
    clip=0.2,
    lam=0.95,
    discount=0.99,
    grad_norm=0.5,
    iterations=100,
    workers=2,
    clip_values=False
)

# DISCRETE: general hp settings that solve
#   - CartPole (turn of state and reward normalization though)
#   - Acrobot
#   - MountainCar (norming is absolutely crucial here

discrete = make_config(
    batch_size=128,
    horizon=2048,
    c_entropy=0.01,
    lr_pi=0.0003,
    epochs=10,
    clip=0.2,
    lam=0.95,
    discount=0.99,
    grad_norm=0.5,
    iterations=100,
    workers=8,
    clip_values=False
)

discrete_no_ent = derive_config(discrete, {"c_entropy": 0.0})
discrete_shared = derive_config(discrete, {"shared": True})
discrete_rnn = derive_config(discrete, {"model": "rnn"})
discrete_gru = derive_config(discrete, {"model": "gru"})
discrete_lstm = derive_config(discrete, {"model": "lstm"})
discrete_no_norms = derive_config(discrete, {"no_state_norming": True, "no_reward_norming": True})
discrete_no_norms_short = derive_config(discrete_no_norms, {"horizon": 512})

# CONTINUOUS DEFAULT
#   - LunarLanderContinuous
#   - BipedalWalker (use beta distribution to crack the reward threshold)

continuous = make_config(
    batch_size=512,
    horizon=2048,
    c_entropy=0.0,
    lr_pi=0.0003,
    epochs=10,
    clip=0.2,
    lam=0.95,
    discount=0.99,
    grad_norm=0.5,
    iterations=100,
    workers=8,
    clip_values=False
)

continuous_shared = derive_config(continuous, {"shared": True})
continuous_rnn = derive_config(continuous, {"model": "rnn"})
continuous_gru = derive_config(continuous, {"model": "gru"})
continuous_lstm = derive_config(continuous, {"model": "lstm"})
continuous_beta = derive_config(continuous, {"distribution": "beta"})
continuous_no_norms = derive_config(continuous, {"no_state_norming": True,
                                                 "no_reward_norming": True})

continuous_with_ent = derive_config(continuous, {"c_entropy": 0.01})
continuous_beta_with_ent = derive_config(continuous_beta, {"c_entropy": 0.01})

continuous_beta_lstm = derive_config(continuous_beta, {"model": "lstm"})
continuous_beta_gru = derive_config(continuous_beta, {"model": "gru"})
continuous_beta_rnn = derive_config(continuous_beta, {"model": "rnn"})

# PENDULUM (for, well, Pendulum-v0)

pendulum = derive_config(continuous, dict(
    horizon=512,
    epochs=10,
    discount=0.99,
    distribution="gaussian"
))

pendulum_beta = derive_config(pendulum, {"distribution": "beta"})

# FROM PAPERS

beta_paper = make_config(
    # continuous with some parameters from the beta paper
    batch_size=64,
    horizon=2048,
    c_entropy=0.001,
    lr_pi=0.0003,
    epochs=10,
    clip=0.2,
    lam=0.95,
    discount=0.995,
    grad_norm=0.5,
    iterations=100,
    workers=8,
    clip_values=False
)

# MUJOCO

mujoco = make_config(
    iterations=1000000 // 2048,  # one million timesteps
    workers=1,
    batch_size=64,
    horizon=2048,
    c_entropy=0.0,
    lr_pi=0.0003,
    epochs=10,
    clip=0.2,
    lam=0.95,
    discount=0.99,
    grad_norm=0.5,
    clip_values=False
)

mujoco_beta = derive_config(mujoco, {"distribution": "beta"})
mujoco_beta_rnn = derive_config(mujoco_beta, {"model": "rnn"})
mujoco_beta_gru = derive_config(mujoco_beta, {"model": "gru"})
mujoco_beta_lstm = derive_config(mujoco_beta, {"model": "lstm"})
mujoco_vc = derive_config(mujoco, {"clip_values": True})
mujoco_beta_shared = derive_config(mujoco_beta, {"shared": True})

# ROBOSCHOOL TASKS

roboschool = make_config(
    iterations=50000000 // (36 * 512),  # 50 million timesteps
    workers=36,
    batch_size=4096,
    horizon=512,
    c_entropy=0.0,
    lr_pi=0.0003,
    lr_schedule="exponential",  # should be a linear annealing
    epochs=15,
    clip=0.2,
    lam=0.95,
    discount=0.99,
    grad_norm=0.5,
    clip_values=False,
    radical_evaluation=True,  # we need to evaluate everytime, because horizon is short than max steps, which biases
    # statistics towards failing episodes
)

roboschool_beta = derive_config(roboschool, {"distribution": "beta"})

# HAND ENVS

hand = make_config(
    iterations=50000,
    workers=16,
    batch_size=256,
    horizon=512,
    c_entropy=0.001,
    lr_pi=5e-4,
    lr_schedule="exponential",
    epochs=10,
    clip=0.2,
    lam=0.95,
    discount=0.998,
    grad_norm=0.5,
    clip_values=False,
    distribution="beta",
    architecture="shadow",
    model="gru"
)

hand_no_ent = derive_config(hand, {"c_entropy": 0.0})
hand_manipulate = derive_config(hand, {
    "c_entropy": 0.01,
    "epochs": 3
})

andrypulate = make_config(
    iterations=50000,
    workers=384,
    batch_size=384 * 32,
    horizon=2048,
    c_entropy=0.001,
    lr_pi=3e-4,
    lr_schedule="exponential",
    epochs=3,
    clip=0.2,
    lam=0.95,
    discount=0.998,
    grad_norm=0.5,
    tbptt=8,
    clip_values=False,
    distribution="multi-categorical",
    architecture="wider",
    model="lstm",
)

akkaypulate = make_config(
    iterations=50000,
    workers=396,   # akkaya et al use 400 but that would not divide by 12; on cscs, this needs 33 gpu nodes
    batch_size=256,
    horizon=4096,
    c_entropy=0.01,
    lr_pi=3e-4,
    lr_schedule="exponential",
    epochs=3,
    clip=0.2,
    lam=0.95,
    discount=0.998,
    grad_norm=0.5,
    clip_values=False,
    distribution="beta",
    architecture="wider",
    model="lstm"
)

manipulate = make_config(
    iterations=50000,
    workers=384,   # need divisibility by 12
    batch_size=384 * 32,
    horizon=2048,
    c_entropy=0.01,
    lr_pi=3e-4,
    lr_schedule="exponential",
    epochs=3,
    clip=0.2,
    lam=0.95,
    discount=0.998,
    grad_norm=0.5,
    clip_values=False,
    distribution="beta",
    architecture="wider",
    model="lstm"
)

manipulate_discrete = derive_config(manipulate, dict(
    distribution="multi-categorical",
    c_entropy=0.001,
))

# RECOMMENDED CONFIGS FOR ENVs

recommended_config = dict(
    **dict.fromkeys(
        ["CartPole-v0", "CartPole-v1"], discrete_no_norms
    ), **dict.fromkeys(
        ["Acrobot-v1", "MountainCar-v0", "LunarLander-v2"], discrete
    ), **dict.fromkeys(
        ["Pendulum-v1"], pendulum_beta
    ), **dict.fromkeys(
        ["LunarLanderContinuous-v2", "BipedalWalker-v3", "BipedalWalkerHardcore-v2"], continuous_beta
    ), **dict.fromkeys(
        ["HalfCheetah-v2", "Hopper-v2", "InvertedPendulum-v2", "InvertedDoublePendulum-v2",
         "Swimmer-v2", "Walker2d-v2"], mujoco_beta
    ), **dict.fromkeys(
        ["Reacher-v2"], mujoco
    ), **dict.fromkeys(
        ["Humanoid-v2", "HumanoidStandup-v2"], roboschool
    ),
)

if __name__ == '__main__':
    import pandas as pd

    config_table = pd.DataFrame.from_dict(
        dict(
            discrete=discrete,
            continuous=continuous,
            pendulum=pendulum,
            mujoco=mujoco,
            roboschool=roboschool,
            hand=hand
        )
    )

    config_table = config_table.drop(
        ["env", "radical_evaluation", "preload", "sequential", "debug", "cpu", "export_file",
         "load_from", "save_every"])

    with open("../../docs/hp_table.tex", "w") as f:
        config_table.to_latex(f, bold_rows=True)



import abc
import sys
from collections import namedtuple
from typing import List, Dict

import numpy as np
import gymnasium as gym

from angorapy.common.senses import Sensation
from angorapy.common.const import NP_FLOAT_PREC, EPSILON
from angorapy.utilities.dtypes import StepTuple
from angorapy.utilities.core import env_extract_dims


PostProcessorSerialization = namedtuple("PostProcessorSerialization", ["class_name", "env_id", "data"])


class BasePostProcessor(abc.ABC):
    """Abstract base class for postprocessors."""

    def __init__(self, env_name: str, state_dim, n_actions):
        self.n = 1e-4  # make this at least epsilon so that first measure is not all zeros
        self.previous_n = self.n

        # extract env info
        self.env_name = env_name
        self.state_dim, self.number_of_actions = state_dim, n_actions

    @abc.abstractmethod
    def __add__(self, other):
        pass

    def __repr__(self):
        return self.__class__.__name__

    @property
    def name(self):
        """The name of the transformers."""
        return self.__class__.__name__

    def transform(self, step_result: StepTuple, **kwargs) -> StepTuple:
        """Transform the step results."""
        pass

    def update(self, **kwargs):
        """Just for the sake of interchangeability, all transformers have an update method even if they do not update."""
        pass

    def warmup(self, env: "TaskWrapper", n_steps=10):
        """Warm up the transformer for n steps."""
        pass

    @abc.abstractmethod
    def serialize(self) -> dict:
        """Serialize the transformer to allow for saving its data in a file."""
        pass

    @staticmethod
    def from_serialization(serialization: PostProcessorSerialization):
        """Create transformer from a serialization."""
        return getattr(sys.modules[__name__], serialization.class_name).recover(serialization.data)

    @classmethod
    @abc.abstractmethod
    def recover(cls, env_id, state_dim, n_actions, serialization_data: list):
        """Recover from serialization."""
        pass

    @staticmethod
    def from_collection(collection_of_transformers):
        """Merge a list of transformers into one new transformer of the same type."""
        assert len(set([type(w) for w in collection_of_transformers])) == 1, \
            "All transformers need to have the same type."

        new_transformer = collection_of_transformers[0]
        for transformer in collection_of_transformers[1:]:
            new_transformer += transformer
        return new_transformer

    def correct_sample_size(self, deduction):
        """Deduce the given number from the sample counter."""
        self.n = self.n - deduction


class BaseRunningMeanPostProcessor(BasePostProcessor, abc.ABC):
    """Abstract base class for transformers implementing a running mean over some statistic."""

    mean: Dict[str, np.ndarray]
    variance: Dict[str, np.ndarray]

    def __add__(self, other) -> "BaseRunningMeanPostProcessor":
        nt = self.__class__(self.env_name, self.state_dim, self.number_of_actions)
        nt.n = self.n + other.n

        for name in nt.mean.keys():
            nt.mean[name] = (self.n / nt.n) * self.mean[name] + (other.n / nt.n) * other.mean[name]
            nt.variance[name] = (self.n * (self.variance[name] + (self.mean[name] - nt.mean[name]) ** 2)
                                 + other.n * (other.variance[name] + (other.mean[name] - nt.mean[name]) ** 2)) / nt.n

        return nt

    def update(self, observation: dict) -> None:
        """Update the mean(serialization) and variance(serialization) of the tracked statistic based on the new sample.

        Simplification of https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm.."""
        self.n += 1

        for name, obs in filter(lambda o: isinstance(o[1], (int, float)) or len(o[1].shape) in [0, 1],
                                observation.items()):
            delta = obs - self.mean[name]
            m_a = self.variance[name] * (self.n - 1)

            self.mean[name] = np.array(self.mean[name] + delta * (1 / self.n), dtype=NP_FLOAT_PREC)
            self.variance[name] = np.array((m_a + np.square(delta) * (self.n - 1) / self.n) / self.n,
                                           dtype=NP_FLOAT_PREC)

    def serialize(self) -> PostProcessorSerialization:
        """Serialize the transformer to allow for saving it in a file."""
        return PostProcessorSerialization(self.__class__.__name__,
                                          self.env_name,
                                          [self.n,
                                         {n: m.tolist() for n, m in self.mean.items()},
                                         {n: v.tolist() for n, v in self.variance.items()}])

    @classmethod
    def recover(cls, env_id, state_dim, n_actions, data: PostProcessorSerialization):
        """Recover a running mean transformer from its serialization"""
        transformer = cls(env_id, state_dim, n_actions)
        transformer.n = np.array(data[0])

        # backwards compatibility
        compatible_data_means = {name if name != "somatosensation" else "touch": l for name, l in data[1].items()}
        compatible_data_variances = {name if name != "somatosensation" else "touch": l for name, l in data[2].items()}
        compatible_data_means = {name if name != "asynchronous" else "asymmetric": l for name, l in compatible_data_means.items()}
        compatible_data_variances = {name if name != "asynchronous" else "asymmetric": l for name, l in compatible_data_variances.items()}

        transformer.mean = {name: np.array(l) for name, l in compatible_data_means.items()}
        transformer.variance = {name: np.array(l) for name, l in compatible_data_variances.items()}

        return transformer

    def simplified_mean(self) -> Dict[str, float]:
        """Get a simplified, one dimensional mean by meaning any means."""
        return {n: np.mean(m).item() for n, m in self.mean.items()}

    def simplified_variance(self) -> Dict[str, float]:
        """Get a simplified, one dimensional variance by meaning any variances."""
        return {n: np.mean(v).item() for n, v in self.variance.items()}

    def simplified_stdev(self) -> Dict[str, float]:
        """Get a simplified, one dimensional stdev by meaning any variances and taking their square root."""
        return {n: np.sqrt(np.mean(v)).item() for n, v in self.variance.items()}


class StateNormalizer(BaseRunningMeanPostProcessor):
    """Transformer for state normalization using running mean and variance estimations."""

    def __init__(self, env_name: str, state_dim, n_actions):
        # parse input types into normed shape format
        super().__init__(env_name, state_dim, n_actions)

        self.shapes = self.state_dim

        self.mean = {k: np.zeros(i_shape, NP_FLOAT_PREC) for k, i_shape in self.shapes.items() if len(i_shape) == 1}
        self.variance = {k: np.ones(i_shape, NP_FLOAT_PREC) for k, i_shape in self.shapes.items() if len(i_shape) == 1}

        assert len(self.mean) > 0 and len(self.variance) > 0, "Initialized StateNormalizationTransformer got no vector " \
                                                              "states."

    def transform(self, step_result: StepTuple, update=True) -> StepTuple:
        """Normalize a given batch of 1D tensors and update running mean and std."""
        o, r, terminated, truncated, info = step_result

        if update:
            self.update(o.dict())

        normed_o = {}
        # for sense_name, sense_value in filter(lambda a: len(a[1].shape) == 1, o.dict().items()):
        for sense_name, sense_value in o.dict().items():
            if len(sense_value.shape) == 1:
                normed_o[sense_name] = np.clip(
                    (sense_value - self.mean[sense_name]) / (np.sqrt(self.variance[sense_name] + EPSILON)), -10., 10.)
            else:
                # assuming visual RGB values
                normed_o[sense_name] = np.divide(sense_value, 255)

        normed_o = Sensation(**normed_o)

        return normed_o, r, terminated, truncated, info

    def warmup(self, env: "TaskWrapper", n_steps=10):
        """Warmup the transformer by sampling the observation space."""
        return  # todo we cannot do this like this, the update will be based on a transformed input this way
        env.reset()
        for i in range(n_steps):
            self.update(env.step(env.action_space.sample())[0].dict())


class RewardNormalizer(BaseRunningMeanPostProcessor):
    """Transformer for reward normalization using running mean and variance estimations."""

    def __init__(self, env_name: str, state_dim, n_action):
        super().__init__(env_name, state_dim, n_action)

        self.mean = {"reward": np.array(0, NP_FLOAT_PREC)}
        self.variance = {"reward": np.array(1, NP_FLOAT_PREC)}
        self.ret = np.float64(0)

    def transform(self, step_tuple: StepTuple, update=True) -> StepTuple:
        """Normalize a given batch of 1D tensors and update running mean and std."""
        o, r, terminated, truncated, info = step_tuple

        if r is None:
            return o, r, terminated, truncated, info  # skip

        # update based on cumulative discounted reward
        if update:
            self.ret = 0.99 * self.ret + r
            self.update({"reward": self.ret})

        # normalize
        r = np.clip(r / (np.sqrt(self.variance["reward"] + EPSILON)), -10., 10.)

        if terminated or truncated:
            self.ret = 0.

        return o, r, terminated, truncated, info

    def warmup(self, env: "TaskWrapper", n_steps=10):
        """Warmup the transformer by randomly stepping the environment through action space sampling."""
        env.reset()
        for i in range(n_steps):
            self.update({"reward": env.step(env.action_space.sample())[1]})


class StateMemory(BasePostProcessor):
    pass


def merge_postprocessors(transformers: List[BasePostProcessor]) -> BasePostProcessor:
    """Merge a list of transformers into a single transformer.

    Args:
        transformers:           list of transformers
    """
    assert all(type(t) is type(transformers[0]) for t in transformers), \
        "To merge postprocessors, they must be of same type."

    # merge the list of transformers into a single transformer
    merged_transformer = BasePostProcessor.from_collection(transformers)

    # adjust for overcounting
    merged_transformer.correct_sample_size((len(transformers) - 1) * transformers[0].previous_n)

    # record the new n for next sync
    merged_transformer.previous_n = merged_transformer.n

    return merged_transformer


def postprocessors_from_serializations(list_of_serializations: List[PostProcessorSerialization]) -> List[BasePostProcessor]:
    """From a list of TransformerSerializations recover the respective transformers."""
    transformers = []
    reference_env = gym.make(PostProcessorSerialization(*list_of_serializations[0]).env_id, render_mode="rgb_array")  # todo could leak memory
    state_dim, n_actions = env_extract_dims(reference_env)

    for cereal in list_of_serializations:
        cereal = PostProcessorSerialization(*cereal)

        class_name = cereal.class_name

        # backward compatibility with formerly used class names
        if class_name == "StateNormalizationTransformer":
            class_name = "StateNormalizer"
        elif class_name == "RewardNormalizationTransformer":
            class_name = "RewardNormalizer"

        transformers.append(
            getattr(sys.modules[__name__], class_name).recover(cereal.env_id, state_dim, n_actions, cereal.data)
        )

    return transformers



import itertools
import json
import os


import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
import matplotlib
import statsmodels.stats.api as sms

from angorapy.common.const import PATH_TO_EXPERIMENTS, QUALITATIVE_COLOR_PALETTE


font = {'family' : 'Times',
        'weight' : 'normal',
        'size'   : 15}

matplotlib.rc('font', **font)

# experiment_ids = [['1675028736765791', '1674983643322591'],
#                   ['1674985163288377', '1674983177286330'],
#                   ['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]  # compare distributions
# names = ["beta", "gaussian", "multicategorical"]

experiment_ids = [['1680268307741971', '1680267170902356', '1680266899101867', '1679765936722940', '1679142835973298'],
                  ['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]
experiment_ids = [['1680268307741971', '1680267170902356', '1679765936722940', '1679142835973298'],
                  ['1674343261520188', '1674308148646663', '1674074113967956', '1673350499432390']]  # compare distributions
names = ["FPN model", "OpenAI"]

# experiment_ids = [['1674975602294059', '1674975602446141', '1671749718306373'],
#                   ['1674343261520188', '1674308148646663', '1674074113967956', '1674074113731734', '1673350499432390']]
# names = ["without auxiliary input", "with auxiliary input"]

cosucc_developments = {}
cosucc_bands = {}

for i, group in enumerate(experiment_ids):
    exp_name = names[i]
    cosucc_developments[exp_name] = []

    for id in group:
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
            meta = json.load(f)
        with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "statistics.json")) as f:
            statistics = json.load(f)

        cosucc_developments[exp_name].append(statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"])

    max_length = max(len(arr) for arr in cosucc_developments[exp_name])

    # pad list of arrays to same length with last value
    cosucc_developments[exp_name] = np.array([np.pad(arr, (0, max_length - len(arr)), mode='edge') for arr in cosucc_developments[exp_name]])

    # cosucc_developments[exp_name] = np.array(
    #     [trace[:min(map(len, cosucc_developments[exp_name]))] for trace in cosucc_developments[exp_name]])

    descm = sms.DescrStatsW(cosucc_developments[exp_name])

    cd_mean = np.mean(cosucc_developments[exp_name], axis=0)
    cosucc_bands[exp_name] = descm.tconfint_mean()

ax1 = plt.subplot2grid((1, 3), (0, 0), colspan=3)
# ax2 = plt.subplot2grid((1, 3), (0, 2), colspan=1)

x_max = len(list(cosucc_developments.items())[0][1][0])
for i, (name, rewards) in enumerate(cosucc_developments.items()):
    x_max = max(x_max, np.max(np.argmax(rewards, axis=-1)))

    mean_line = np.mean(cosucc_developments[name], axis=0)
    argmax_mean_line = np.argmax(mean_line)

    mean_line = mean_line[:np.argmax(mean_line)]
    ax1.plot(mean_line, label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    ax1.fill_between(range(len(cosucc_bands[name][0]))[:argmax_mean_line],
                     cosucc_bands[name][0][:argmax_mean_line],
                     cosucc_bands[name][1][:argmax_mean_line],
                     alpha=0.2)

    # plot horizontal line indicating max performance with circles as markers
    ax1.axhline(y=np.max(mean_line), color=QUALITATIVE_COLOR_PALETTE[i], linestyle="--", linewidth=1)

    # plot horizontal line indicating max performance of best agent in group
    ax1.axhline(y=np.max(np.max(cosucc_developments[name], axis=0)), linestyle=":", color=QUALITATIVE_COLOR_PALETTE[i], linewidth=0.5)

    # ax2.hist(
    #     evaluation_reward_histograms[name][1][:-1],
    #     evaluation_reward_histograms[name][1],
    #     weights=evaluation_reward_histograms[name][0],
    #     edgecolor="white", linewidth=0.5, alpha=0.8
    # )
    # ax2.set_xlabel("Consecutive Goals Reached")
    # ax2.set_ylabel("Number of Episodes")

# df = pd.DataFrame(
#     {"group": itertools.chain(*[[name] * len(dp) for name, dp in evaluation_rewards.items()]),
#      "Consecutive Goals Reached": np.concatenate([dp for name, dp in evaluation_rewards.items()])}
# )

# sns.boxplot(data=df, x="group", y="Consecutive Goals Reached", medianprops={"color": "red"}, flierprops={"marker": "x"},
#             fliersize=1, ax=ax2, palette={name: QUALITATIVE_COLOR_PALETTE[i] for i, name in enumerate(names)},
#             showmeans=True, meanprops={"marker": "o", "markerfacecolor": "white", "markeredgecolor": "black"})
# ax2.set_xlabel("")

ax1.set_xlabel("Cycle")
ax1.set_ylabel("Consecutive Goals Reached")

# legend with shorter lines but normal font
ax1.legend(loc='upper left',  ncol=1, handlelength=.7)

ax1.set_xlim(0, x_max)
ax1.set_ylim(0, 50)
ax1.set_yticks(np.arange(0, 50, 10))
ax1.set_xticks(np.arange(0, 1600, 200))

plt.gcf().set_size_inches(8, 4)
plt.subplots_adjust(wspace=0., right=0.999, left=0.1, top=0.999, bottom=0.15)
# plt.show()
plt.savefig("../../../docs/figures/brain-vs-openai.png", format="png")



import random

import numpy as np


def generate_random_sim_qpos(base: dict) -> dict:
    """Generate a random state of the simulation."""
    for key, val in base.items():
        if key in ["robot/WRJ1", "robot/WRJ0"]:
            continue  # do not randomize the wrist

        base[key] = val + val * random.gauss(0, 0.1)

    return base


def get_fingertip_distance(ft_a, ft_b):
    """Return the distance between two vectors representing finger tip positions."""
    assert ft_a.shape == ft_b.shape
    return np.linalg.norm(ft_a - ft_b, axis=-1)


def quat_from_angle_and_axis(angle, axis):
    assert axis.shape == (3,)
    axis /= np.linalg.norm(axis)
    quat = np.concatenate([[np.cos(angle / 2.)], np.sin(angle / 2.) * axis])
    quat /= np.linalg.norm(quat)
    return quat



#!/usr/bin/env python
"""Custom data types for simplifying data communication."""
import itertools
import statistics
from collections import namedtuple
from typing import List, Union

import numpy as np

try:
    from mpi4py import MPI
except ImportError:
    MPI = None


StatBundle = namedtuple("StatBundle", ["numb_completed_episodes", "numb_processed_frames",
                                       "episode_rewards", "episode_lengths", "tbptt_underflow",
                                       "per_receptor_mean", "auxiliary_performances"])


def condense_stats(stat_bundles: List[StatBundle]) -> Union[StatBundle, None]:
    """Infer a single StatBundle from a list of StatBundles."""
    return StatBundle(
        numb_completed_episodes=sum([s.numb_completed_episodes for s in stat_bundles]),
        numb_processed_frames=sum([s.numb_processed_frames for s in stat_bundles]),
        episode_rewards=list(itertools.chain(*[s.episode_rewards for s in stat_bundles])),
        episode_lengths=list(itertools.chain(*[s.episode_lengths for s in stat_bundles])),
        tbptt_underflow=round(statistics.mean(map(lambda x: x.tbptt_underflow, stat_bundles)), 2) if (
                stat_bundles[0].tbptt_underflow is not None) else None,
        per_receptor_mean={
            sense: np.mean([s.per_receptor_mean[sense] for s in stat_bundles], axis=0)
            for sense in stat_bundles[0].per_receptor_mean.keys()
        },
        auxiliary_performances={
            aux: list(itertools.chain(*[s.auxiliary_performances[aux] for s in stat_bundles]))
            for aux in stat_bundles[0].auxiliary_performances.keys()
        }
    )


def mpi_condense_stats(stat_bundles: List[StatBundle]) -> Union[StatBundle, None]:
    """Pull together the StatBundle lists from the buffer on all workers and infer a single StatBundle."""
    if MPI is not None:
        stat_bundles = MPI.COMM_WORLD.gather(stat_bundles, root=0)
        rank = MPI.COMM_WORLD.Get_rank()
    else:
        stat_bundles = [stat_bundles]
        rank = 0

    if rank == 0:
        stat_bundles = list(itertools.chain(*stat_bundles))
        return condense_stats(stat_bundles)
    else:
        return None



import numpy as np
import tensorflow as tf
from scipy.integrate import odeint
from scipy.special import binom


def library_size(n, poly_order, use_sine=False, include_constant=True):
    l = 0
    for k in range(poly_order + 1):
        l += int(binom(n + k - 1, k))
    if use_sine:
        l += n
    if not include_constant:
        l -= 1
    return l


def sindy_fit(RHS, LHS, coefficient_threshold):
    m, n = LHS.shape
    Xi = np.linalg.lstsq(RHS, LHS, rcond=None)[0]

    for k in range(10):
        small_inds = (np.abs(Xi) < coefficient_threshold)
        Xi[small_inds] = 0
        for i in range(n):
            big_inds = ~small_inds[:, i]
            if np.where(big_inds)[0].size == 0:
                continue
            Xi[big_inds, i] = np.linalg.lstsq(RHS[:, big_inds], LHS[:, i], rcond=None)[0]
    return Xi


def sindy_library_tf(z, latent_dim, poly_order, include_sine=False):
    """
    Build the SINDy library.
    Arguments:
        z - 2D tensorflow array of the snapshots on which to build the library. Shape is number of
        time points by the number of state variables.
        latent_dim - Integer, number of state variable in z.
        poly_order - Integer, polynomial order to which to build the library. Max value is 5.
        include_sine - Boolean, whether or not to include sine terms in the library. Default False.
    Returns:
        2D tensorflow array containing the constructed library. Shape is number of time points by
        number of library functions. The number of library functions is determined by the number
        of state variables of the input, the polynomial order, and whether or not sines are included.
    """
    library = [tf.ones(tf.shape(z)[0])]

    for i in range(latent_dim):
        library.append(z[:, i])

    if poly_order > 1:
        for i in range(latent_dim):
            for j in range(i, latent_dim):
                library.append(tf.multiply(z[:, i], z[:, j]))

    if poly_order > 2:
        for i in range(latent_dim):
            for j in range(i, latent_dim):
                for k in range(j, latent_dim):
                    library.append(z[:, i] * z[:, j] * z[:, k])

    if poly_order > 3:
        for i in range(latent_dim):
            for j in range(i, latent_dim):
                for k in range(j, latent_dim):
                    for p in range(k, latent_dim):
                        library.append(z[:, i] * z[:, j] * z[:, k] * z[:, p])

    if poly_order > 4:
        for i in range(latent_dim):
            for j in range(i, latent_dim):
                for k in range(j, latent_dim):
                    for p in range(k, latent_dim):
                        for q in range(p, latent_dim):
                            library.append(z[:, i] * z[:, j] * z[:, k] * z[:, p] * z[:, q])

    if include_sine:
        for i in range(latent_dim):
            library.append(tf.sin(z[:, i]))

    return tf.stack(library, axis=1)


def sindy_library_tf_order2(z, dz, latent_dim, poly_order, include_sine=False):
    """
    Build the SINDy library for a second order system. This is essentially the same as for a first
    order system, but library terms are also built for the derivatives.
    """
    library = [tf.ones(tf.shape(z)[0])]

    z_combined = tf.concat([z, dz], 1)

    for i in range(2 * latent_dim):
        library.append(z_combined[:, i])

    if poly_order > 1:
        for i in range(2 * latent_dim):
            for j in range(i, 2 * latent_dim):
                library.append(tf.multiply(z_combined[:, i], z_combined[:, j]))

    if poly_order > 2:
        for i in range(2 * latent_dim):
            for j in range(i, 2 * latent_dim):
                for k in range(j, 2 * latent_dim):
                    library.append(z_combined[:, i] * z_combined[:, j] * z_combined[:, k])

    if poly_order > 3:
        for i in range(2 * latent_dim):
            for j in range(i, 2 * latent_dim):
                for k in range(j, 2 * latent_dim):
                    for p in range(k, 2 * latent_dim):
                        library.append(z_combined[:, i] * z_combined[:, j] * z_combined[:, k] * z_combined[:, p])

    if poly_order > 4:
        for i in range(2 * latent_dim):
            for j in range(i, 2 * latent_dim):
                for k in range(j, 2 * latent_dim):
                    for p in range(k, 2 * latent_dim):
                        for q in range(p, 2 * latent_dim):
                            library.append(
                                z_combined[:, i] * z_combined[:, j] * z_combined[:, k] * z_combined[:, p] * z_combined[
                                                                                                            :, q])

    if include_sine:
        for i in range(2 * latent_dim):
            library.append(tf.sin(z_combined[:, i]))

    return tf.stack(library, axis=1)


def compute_z_derivatives(input, dx, weights, biases):
    """
    Compute the first order time derivatives by propagating through the network.
    Arguments:
        input - 2D tensorflow array, input to the network. Dimensions are number of time points
        by number of state variables.
        dx - First order time derivatives of the input to the network.
        weights - List of tensorflow arrays containing the network weights
        biases - List of tensorflow arrays containing the network biases
        activation - String specifying which activation function to use. Options are
        'elu' (exponential linear unit), 'relu' (rectified linear unit), 'sigmoid',
        or linear.
    Returns:
        dz - Tensorflow array, first order time derivatives of the network output.
    """
    dz = dx
    for i in range(len(weights) - 1):
        input = tf.matmul(input, weights[i]) + biases[i]
        dz = tf.multiply(tf.cast(input > 0, tf.float32), tf.matmul(dz, weights[i]))
        input = tf.nn.relu(input)
    dz = tf.matmul(dz, weights[-1])

    return dz


def z_derivative_order2(input, dx, ddx, weights, biases, activation='elu'):
    """
    Compute the first and second order time derivatives by propagating through the network.
    Arguments:
        input - 2D tensorflow array, input to the network. Dimensions are number of time points
        by number of state variables.
        dx - First order time derivatives of the input to the network.
        ddx - Second order time derivatives of the input to the network.
        weights - List of tensorflow arrays containing the network weights
        biases - List of tensorflow arrays containing the network biases
        activation - String specifying which activation function to use. Options are
        'elu' (exponential linear unit), 'relu' (rectified linear unit), 'sigmoid',
        or linear.
    Returns:
        dz - Tensorflow array, first order time derivatives of the network output.
        ddz - Tensorflow array, second order time derivatives of the network output.
    """
    dz = dx
    ddz = ddx
    if activation == 'elu':
        for i in range(len(weights) - 1):
            input = tf.matmul(input, weights[i]) + biases[i]
            dz_prev = tf.matmul(dz, weights[i])
            elu_derivative = tf.minimum(tf.exp(input), 1.0)
            elu_derivative2 = tf.multiply(tf.exp(input), tf.to_float(input < 0))
            dz = tf.multiply(elu_derivative, dz_prev)
            ddz = tf.multiply(elu_derivative2, tf.square(dz_prev)) \
                  + tf.multiply(elu_derivative, tf.matmul(ddz, weights[i]))
            input = tf.nn.elu(input)
        dz = tf.matmul(dz, weights[-1])
        ddz = tf.matmul(ddz, weights[-1])
    elif activation == 'relu':
        # NOTE: currently having trouble assessing accuracy of 2nd derivative due to discontinuity
        for i in range(len(weights) - 1):
            input = tf.matmul(input, weights[i]) + biases[i]
            relu_derivative = tf.to_float(input > 0)
            dz = tf.multiply(relu_derivative, tf.matmul(dz, weights[i]))
            ddz = tf.multiply(relu_derivative, tf.matmul(ddz, weights[i]))
            input = tf.nn.relu(input)
        dz = tf.matmul(dz, weights[-1])
        ddz = tf.matmul(ddz, weights[-1])
    elif activation == 'sigmoid':
        for i in range(len(weights) - 1):
            input = tf.matmul(input, weights[i]) + biases[i]
            input = tf.sigmoid(input)
            dz_prev = tf.matmul(dz, weights[i])
            sigmoid_derivative = tf.multiply(input, 1 - input)
            sigmoid_derivative2 = tf.multiply(sigmoid_derivative, 1 - 2 * input)
            dz = tf.multiply(sigmoid_derivative, dz_prev)
            ddz = tf.multiply(sigmoid_derivative2, tf.square(dz_prev)) \
                  + tf.multiply(sigmoid_derivative, tf.matmul(ddz, weights[i]))
        dz = tf.matmul(dz, weights[-1])
        ddz = tf.matmul(ddz, weights[-1])
    else:
        for i in range(len(weights) - 1):
            dz = tf.matmul(dz, weights[i])
            ddz = tf.matmul(ddz, weights[i])
        dz = tf.matmul(dz, weights[-1])
        ddz = tf.matmul(ddz, weights[-1])
    return dz, ddz


if __name__ == '__main__':
    print(sindy_library_tf(tf.random.normal((1024, 100), dtype=tf.float32), latent_dim=50, poly_order=2))



"""interactive plotting from https://stackoverflow.com/a/31417070/5407682"""
import colorsys
import math
from typing import List

import numpy
import numpy as np
from matplotlib import pyplot as plt, colors as mc


def make_interactive_legend(ax=None):
    if ax is None:
        ax = plt.gca()
    if ax.legend_ is None:
        ax.legend()

    return InteractiveLegend(ax.get_legend())


class InteractiveLegend(object):
    def __init__(self, legend):
        self.legend = legend
        self.fig = legend.axes.figure

        self.lookup_artist, self.lookup_handle = self._build_lookups(legend)
        self._setup_connections()

        self.update()

    def _setup_connections(self):
        for artist in self.legend.texts + self.legend.legendHandles:
            artist.set_picker(10)  # 10 points tolerance

        self.fig.canvas.mpl_connect('pick_event', self.on_pick)
        self.fig.canvas.mpl_connect('button_press_event', self.on_click)

    def _build_lookups(self, legend):
        labels = [t.get_text() for t in legend.texts]
        handles = legend.legendHandles
        label2handle = dict(zip(labels, handles))
        handle2text = dict(zip(handles, legend.texts))

        lookup_artist = {}
        lookup_handle = {}
        for artist in legend.axes.get_children():
            if artist.get_label() in labels:
                handle = label2handle[artist.get_label()]
                lookup_handle[artist] = handle
                lookup_artist[handle] = artist
                lookup_artist[handle2text[handle]] = artist

        lookup_handle.update(zip(handles, handles))
        lookup_handle.update(zip(legend.texts, handles))

        return lookup_artist, lookup_handle

    def on_pick(self, event):
        handle = event.artist
        if handle in self.lookup_artist:
            artist = self.lookup_artist[handle]
            artist.set_visible(not artist.get_visible())
            self.update()

    def on_click(self, event):
        if event.button == 3:
            visible = False
        elif event.button == 2:
            visible = True
        else:
            return

        for artist in self.lookup_artist.values():
            artist.set_visible(visible)
        self.update()

    def update(self):
        for artist in self.lookup_artist.values():
            handle = self.lookup_handle[artist]
            if artist.get_visible():
                handle.set_visible(True)
            else:
                handle.set_visible(False)
        self.fig.canvas.draw()

    def show(self):
        plt.show()


def plot_with_confidence(x, lb, ub, label, col=None, alpha=0.2):
    """Plot a line with confidence Intervals."""
    x = np.array(x)
    plt.fill_between(range(x.shape[0]), lb, ub, alpha=alpha, color=col)
    plt.plot(x, color=col, label=f"{label}")


def lighten_color(color, amount=0.5):
    """Lightens the given color by multiplying (1-luminosity) by the given amount."""
    try:
        c = mc.cnames[color]
    except Exception:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))

    return colorsys.hls_to_rgb(c[0], 1 - amount * (1 - c[1]), c[2])


def plot_image_tiling(images: List[numpy.ndarray], cmap: str = None):
    # prepare subplots
    n_filters = len(images)
    tiles_per_row = math.ceil(math.sqrt(n_filters))
    fig, axes = plt.subplots(tiles_per_row, tiles_per_row)
    plt.subplots_adjust(wspace=0.05, hspace=0.05)

    i = 0
    axes = [[axes]] if not isinstance(axes, numpy.ndarray) else axes
    for row_of_axes in axes:
        for axis in row_of_axes:
            if i < n_filters:
                axis.imshow(images[i], cmap=cmap) if cmap is not None else axis.imshow(images[i])
            else:
                axis.axis("off")
            axis.set_xticks([])
            axis.set_yticks([])
            i += 1

    return fig, axes


def transparent_cmap(cmap, N=255):
    """https://stackoverflow.com/questions/42481203/heatmap-on-top-of-image"""
    cmap._init()
    cmap._lut[:, -1] = numpy.linspace(0, 0.8, N + 4)
    return cmap



from typing import Iterable

import numpy as np
import tensorflow as tf

from angorapy.analysis.util.sindy import compute_z_derivatives, sindy_library_tf, library_size


def build_encoder(layer_sizes: Iterable[int]):
    encoder = tf.keras.Sequential([
        tf.keras.layers.Dense(size, activation="relu") for size in layer_sizes
    ])

    return encoder


def build_decoder(layer_sizes: Iterable[int]):
    decoder = tf.keras.Sequential([
        tf.keras.layers.Dense(size, activation="relu" if i != (len(layer_sizes) - 1) else None) for i, size in
        enumerate(layer_sizes)
    ])

    return decoder


def build_full_autoencoder(layer_sizes: Iterable[int], original_dim: int):
    input_layer = tf.keras.layers.Input((original_dim,))

    encoded = build_encoder(layer_sizes)(input_layer)
    decoded = build_decoder(layer_sizes[1:] + [original_dim])(encoded)

    model = tf.keras.Model(inputs=[input_layer], outputs=[encoded, decoded])

    return model


class SindyAutoencoder(tf.keras.Model):

    def __init__(self, layer_sizes: Iterable[int], z_dim, original_dim: int, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.z_dim = z_dim

        # autoencoder
        self.encoder = build_encoder(layer_sizes)
        self.decoder = build_decoder(layer_sizes[1:] + [original_dim])

        # sindy
        self.coefficients = tf.Variable(
            initial_value=tf.random_normal_initializer()(shape=(library_size(self.z_dim, 2), self.z_dim)),
            trainable=True
        )

    def call(self, inputs, training=None, mask=None):
        z_coordinates = self.encoder(inputs)
        x_reconstruction = self.decoder(z_coordinates)

        z_library = sindy_library_tf(z_coordinates, self.z_dim, poly_order=2)
        z_derivative_approximation = tf.matmul(z_library, self.coefficients)

        return z_coordinates, x_reconstruction, z_derivative_approximation



from pathlib import Path

from setuptools import find_packages
from setuptools import setup

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='angorapy',
    version='0.10.7',
    description='Build Goal-driven Models of the Sensorimotor Cortex with Ease.',
    long_description_content_type='text/markdown',
    long_description=long_description,
    url='http://www.angorapy.org',
    author='Tonio Weidler',
    author_email='research@tonioweidler.de',
    license='GPL-3.0',
    packages=find_packages(),
    install_requires=[
        "imageio==2.28.1",

        # tensorflow and extensions
        "tensorflow==2.12.0",
        "tensorflow_probability==0.20.0",
        "tensorflow_graphics==2021.12.3",
        "tensorflow_graphics==2021.12.3",

        "tqdm",
        "simplejson",
        "psutil",
        "scipy",
        "argcomplete",
        "matplotlib",
        "scikit-learn==1.2.2",
        "pandas==1.4.4",
        "nvidia-ml-py3",
        "seaborn",
        "distance",
        "statsmodels==0.14.0",
        "keras_cortex==0.0.8",
        "mediapy",

        # environments
        "gymnasium[mujoco]==0.28.1",
        "mujoco",
        "dm_control==1.0.12",
        "mujoco_utils",
        "tensorflow_datasets",
    ],

    extras_require={
        "box2d": ["box2d-py==2.3.5",
                  "gymnasium[box2d]==0.28.1"],
        "webinterface": ["itsdangerous==2.0.1",
                         "werkzeug==2.0.3",
                         "Flask~=1.1.2",
                         "Jinja2==3.0.0",
                         "bokeh==2.3.3",
                         "flask_jsglue"],
    },

    include_package_data=True,
    package_data={
        "angorapy": ["tasks/**/*"],
    },

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    python_requires=">=3.8",
)



import os
from typing import List

import matplotlib.image as mpimg
import matplotlib.pyplot as plt
import tensorflow as tf
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from analysis.visualization import Visualizer
from utilities.plotting import transparent_cmap

os.chdir("../../../")

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['CUDA_VISIBLE_DEVICES'] = '-1'

tf.random.set_seed(69420)

# load the input image
reference = mpimg.imread("img/hand.png")
reference_gray = tf.squeeze(tf.image.rgb_to_grayscale(reference)).numpy()
hm_cmap = transparent_cmap(plt.cm.get_cmap("hot"))

# ANALYZERS
analyzer_c = Visualizer.from_saved_model("storage/pretrained/caltech_c.h5", mode="return")
analyzer_r = Visualizer.from_saved_model("storage/pretrained/visual_r.h5", mode="return")
analyzer_p = Visualizer.from_saved_model("storage/pretrained/hands_h.h5", mode="return")

# FEATURE MAPS
fm_heat_c = analyzer_c.feature_map("conv2d_3", reference, mode="heat")
fm_heat_r = analyzer_r.feature_map("conv2d_3", reference, mode="heat")
fm_heat_p = analyzer_p.feature_map("conv2d_3", reference, mode="heat")

plt.close()

axs: List[Axes]
fig: Figure = plt.figure(figsize=(12, 3))
grid = plt.GridSpec(2, 8)

reference_ax = fig.add_subplot(grid[:, :2])
reference_ax.imshow(reference)
reference_ax.set_xticks([]), reference_ax.set_yticks([])

filter_ids = [[10, 14], [17, 27]]

ax: Axes
for row in [0, 1]:
    row_filter_ids = filter_ids[row]
    for col in [2, 3, 4, 5, 6, 7]:
        fi = row_filter_ids[col % 2]
        fm = fm_heat_c
        t_color = "red"
        if col > 3:
            fm = fm_heat_r
            t_color = "green"
        if col > 5:
            fm = fm_heat_p
            t_color = "blue"

        ax = fig.add_subplot(grid[row, col])
        ax.set_facecolor(t_color)
        ax.imshow(reference_gray, cmap="gray")
        ax.contourf(fm[:, :, fi], cmap=hm_cmap)
        ax.set_xticks([])
        ax.set_yticks([])

plt.savefig("docs/figures/pretraining_methods_fms.pdf", format="pdf", bbox_inches='tight')
plt.show()



from typing import List, Tuple, Iterable

import tensorflow as tf

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

try:
    ADAM_BASE = tf.keras.optimizers.legacy.Adam
except:
    ADAM_BASE = tf.keras.optimizers.Adam


class AdamSynchronizationError(Exception):
    pass


class MpiAdam(ADAM_BASE):
    """Adam optimizer that averages gradients across mpi processes."""

    def __init__(self, comm, learning_rate=0.001, epsilon=1e-7):
        super().__init__(learning_rate=learning_rate, epsilon=epsilon)
        self.comm = comm
        self.comm_size = comm.Get_size() if comm is not None else 1
        self.comm_rank = comm.Get_rank() if comm is not None else 0

    def _sync_parameters(self):
        """Syncs all Adam parameters across all MPI processes."""
        if MPI is None:
            root_variables = self.comm.bcast(self.weights(), int_root=0)
        else:
            root_variables = self.weights()
        self.set_weights(root_variables)

    def apply_gradients(self, grads_and_vars: Iterable[Tuple[tf.Tensor, tf.Variable]], name=None, **kwargs):
        """ Apply the gradients after averaging over processes."""
        if self.comm is not None and self.comm.size > 1:
            grads_and_vars = [(self.comm.allreduce(g, op=MPI.SUM) / self.comm.Get_size(), v)
                              for g, v in grads_and_vars]

        context = super().apply_gradients(grads_and_vars)

        # wait for all processes before continuing
        if self.comm is not None:
            self.comm.Barrier()

        return context

    def validate_synced_parameters(self):
        """Validate if all optimization processes share the same Adam parameters (moments)."""
        if self.comm is None:
            return

        all_variables = self.comm.gather(self.weights())
        is_in_sync = all(elem == all_variables[0] for elem in all_variables)

        if not is_in_sync:
            raise AdamSynchronizationError("Parameters of MPIAdam optimizers are out of sync.")

    def serialize(self):
        """Convert the Optimizer into a JSON-compatible serialization for saving."""
        config = self.get_config()
        for n, e in config.items():
            if hasattr(e, "item"):
                config[n] = e.item()

        return config

    @staticmethod
    def from_serialization(comm, serialization, var_list) -> "MpiAdam":
        if isinstance(serialization["learning_rate"], dict):
            schedule_type = serialization["learning_rate"]["class_name"]
            schedule_config = serialization["learning_rate"]["config"]
            learning_rate = getattr(tf.keras.optimizers.schedules, schedule_type).from_config(schedule_config)
        elif isinstance(serialization["learning_rate"], float):
            learning_rate = serialization["learning_rate"]
        else:
            raise ValueError("Unknown learning rate/schedule format.")

        adam = MpiAdam(
            comm=comm,
            learning_rate=learning_rate,
            epsilon=serialization["epsilon"]
        )

        # adam._create_all_weights(var_list)
        adam.apply_gradients(zip([tf.zeros_like(v) for v in var_list], var_list))
        adam.set_weights([tf.convert_to_tensor(v) for v in serialization["weights"][:len(adam.variables())]])  # todo remove the slicing hack; only for backwards compatibility

        return adam



#!/usr/bin/env python
"""Methods for creating a story about a training process."""
import datetime
import os
import socket
from inspect import getfullargspec as fargs

import matplotlib.pyplot as plt
import numpy
import simplejson as json
import tensorflow as tf
from gymnasium.spaces import Box
from matplotlib import animation

try:
    from mpi4py import MPI
except ImportError:
    MPI = None

from angorapy.agent.ppo_agent import PPOAgent
from angorapy.common import const
from angorapy.common.const import PATH_TO_EXPERIMENTS
from angorapy.common.postprocessors import RewardNormalizer
from angorapy.common.postprocessors import StateNormalizer
from angorapy.models import get_model_type
from angorapy.tasks.wrappers import TaskWrapper
from angorapy.utilities.core import add_state_dims
from angorapy.utilities.core import flatten


def scale(vector):
    """Min Max scale a vector."""
    divisor = max(vector) - min(vector)
    return (numpy.array(vector) - min(vector)) / (divisor if divisor != 0 else const.EPSILON)


class Monitor:
    """Monitor for learning progress. Tracks and writes statistics to be parsed by the Flask app."""

    def __init__(self,
                 agent: PPOAgent,
                 env: TaskWrapper = None,
                 frequency: int = 1,
                 gif_every: int = 0,
                 id=None,
                 iterations=None,
                 config_name: str = "unknown",
                 experiment_group: str = "default"):
        self.agent = agent
        self.env = env if env is not None else agent.env
        self.iterations = iterations

        self.experiment_group = experiment_group
        self.config_name = config_name

        self.frequency = frequency
        self.gif_every = gif_every
        self.continuous_control = isinstance(self.env.action_space, Box)

        if id is None:
            self.story_id = agent.agent_id
        else:
            self.story_id = id

        self.story_directory = f"{PATH_TO_EXPERIMENTS}/{self.story_id}/"
        os.makedirs(self.story_directory, exist_ok=True)

        try:
            tf.keras.utils.plot_model(self.agent.joint, to_file=f"{self.story_directory}/model.png", expand_nested=True,
                                      show_shapes=True, dpi=300)
        except:
            print("Could not create model plot.")

        self.make_metadata()
        self.write_progress()

    def create_episode_gif(self, n: int):
        """Make n GIFs with the current policy."""

        # rebuild model with batch size of 1
        pi, _, _ = self.agent.model_builder(self.env,
                                            **({"bs": 1} if "bs" in fargs(self.agent.model_builder).args else {}))
        pi.set_weights(self.agent.policy.get_weights())

        for j in range(n):
            episode_letter = chr(97 + j)

            # collect an episode
            done = False
            frames = []
            state = self.env.reset()
            while not done:
                frames.append(self.env.render(mode="rgb_array"))

                probabilities = flatten(pi.predict(add_state_dims(state, dims=2 if self.agent.is_recurrent else 1)))
                action, _ = self.agent.distribution.act(*probabilities)
                observation, reward, done, _ = self.env.step(
                    numpy.atleast_1d(action) if self.continuous_control else action)
                state = observation

            # the figure
            plt.figure(figsize=(frames[0].shape[1] / 72.0, frames[0].shape[0] / 72.0), dpi=72)
            patch = plt.imshow(frames[0], cmap="Greys" if len(frames[0].shape) == 2 else None)
            plt.axis('off')

            def _animate(i):
                patch.set_data(frames[i])

            anim = animation.FuncAnimation(plt.gcf(), _animate, frames=len(frames), interval=50)
            anim.save(f"{self.story_directory}/iteration_{self.agent.iteration}_{episode_letter}.gif",
                      writer='pillow',
                      fps=25)

            plt.close()

    def make_metadata(self, additional_hps: dict = None):
        """Write meta data information about experiment into json file."""

        if additional_hps is None:
            additional_hps = {}

        metadata = dict(
            date=str(datetime.datetime.now()).split(".")[0],
            config=self.config_name,
            host=socket.gethostname(),
            n_cpus=MPI.COMM_WORLD.size if MPI is not None else 1,
            n_gpus=self.agent.n_optimizers,
            angorapy_version=None,  # pkg_resources.get_distribution("angorapy").version,
            experiment_group=self.experiment_group,
            iterations=self.iterations,
            environment=dict(
                name=self.agent.env_name,
                action_space=str(self.agent.env.action_space),
                observation_space=str(self.agent.env.observation_space),
                deterministic=str(self.agent.env.spec.nondeterministic),
                max_steps=str(self.agent.env.spec.max_episode_steps),
                reward_threshold=str(self.agent.env.spec.reward_threshold),
                max_action_values=str(self.agent.env.action_space.high) if hasattr(self.agent.env.action_space,
                                                                                   "high") else "None",
                min_action_values=str(self.agent.env.action_space.low) if hasattr(self.agent.env.action_space,
                                                                                  "low") else "None",
            ),
            hyperparameters=dict(
                continuous=str(self.agent.continuous_control),
                distribution=self.agent.distribution.short_name,
                model=get_model_type(self.agent.model_builder),
                learning_rate=str(self.agent.learning_rate),
                epsilon_clip=str(self.agent.clip),
                entropy_coefficient=str(self.agent.c_entropy.numpy().item()),
                value_coefficient=str(self.agent.c_value.numpy().item()),
                horizon=str(self.agent.horizon),
                workers=str(self.agent.n_workers),
                discount=str(self.agent.discount),
                GAE_lambda=str(self.agent.lam),
                gradient_clipping=str(self.agent.gradient_clipping),
                clip_values=str(self.agent.clip_values),
                reward_norming=str(any([isinstance(t, RewardNormalizer) for t in self.env.transformers])),
                state_norming=str(any([isinstance(t, StateNormalizer) for t in self.env.transformers])),
                TBPTT_sequence_length=str(self.agent.tbptt_length),
                architecture=self.agent.builder_function_name,
                gatherer=str(self.agent.gatherer_class),
                **additional_hps
            ),
            reward_function=dict(self.agent.env.reward_config if hasattr(self.agent.env, "reward_config") else dict(),
                                 identifier=self.agent.reward_configuration)
        )

        with open(f"{self.story_directory}/meta.json", "w") as f:
            json.dump(metadata, f, ignore_nan=True)

    def write_progress(self):
        """Write progress statistics into json file."""
        progress = dict(
            rewards=dict(
                mean=[round(v, 2) if v is not None else v for v in self.agent.cycle_reward_history],
                stdev=[round(v, 2) if v is not None else v for v in self.agent.cycle_reward_std_history],
                last_cycle=self.agent.episode_reward_history[-1] if self.agent.iteration > 1 else []),
            lengths=dict(
                mean=[round(v, 2) if v is not None else v for v in self.agent.cycle_length_history],
                stdev=[round(v, 2) if v is not None else v for v in self.agent.cycle_length_std_history],
                last_cycle=self.agent.episode_length_history[-1] if self.agent.iteration > 1 else []),
            entropies=[round(v, 4) if v is not None else v for v in self.agent.entropy_history],
            vloss=[round(v, 4) if v is not None else v for v in self.agent.value_loss_history],
            ploss=[round(v, 4) if v is not None else v for v in self.agent.policy_loss_history],
            preprocessors=self.agent.wrapper_stat_history
        )

        with open(f"{self.story_directory}/progress.json", "w") as f:
            json.dump(progress, f, ignore_nan=True)

    def write_statistics(self):
        """Write general statistics into json file."""
        stat_dict = dict(
            training=dict(
                avg_seconds_per_cycle=str(numpy.mean(self.agent.cycle_timings)) if self.agent.cycle_timings else "N/A",
                total_train_time=str(numpy.sum(self.agent.cycle_timings)) if self.agent.cycle_timings else "0",
            ),
            used_memory=self.agent.used_memory,
            used_gpu_memory=self.agent.used_gpu_memory,
            cycle_timings=self.agent.cycle_timings,
            optimization_timings=self.agent.optimization_timings,
            gathering_timings=self.agent.gathering_timings,
            loaded_at=self.agent.loading_history,
            per_receptor_mean=self.agent.current_per_receptor_mean,
            auxiliary_performances=self.agent.auxiliary_performances
        )

        with open(f"{self.story_directory}/statistics.json", "w") as f:
            json.dump(stat_dict, f, ignore_nan=True)

    def update(self):
        """Update different components of the Monitor."""
        self.write_progress()
        self.write_statistics()



from bokeh import colors

from angorapy.common.const import COLORMAP

palette = [colors.RGB(*[int(c * 255) for c in color]) for color in COLORMAP]
darker_palette = [c.darken(0.3) for c in palette]
tooltip_css = """
tooltip {
    background-color: #212121;
    color: white;
    padding: 5px;
    border-radius: 10px;
    margin-left: 10px;
}
"""
plot_styling = dict(
    plot_height=500,
    sizing_mode="stretch_width",
    toolbar_location="above",
    tools=""
)


def style_plot(p):
    p.xgrid.visible = False
    p.ygrid.visible = False
    p.outline_line_color = None

    p.axis.axis_label_text_font = "times"
    p.axis.axis_label_text_font_size = "12pt"
    p.axis.axis_label_text_font_style = "bold"

    p.legend.label_text_font = "times"
    p.legend.label_text_font_size = "12pt"
    p.legend.label_text_font_style = "normal"

    p.title.align = "center"
    p.title.text_font_size = "14pt"
    p.title.text_font = "Fira Sans"

    p.toolbar.logo = None


import json
import os

import numpy as np
from matplotlib import pyplot as plt

from angorapy.common.const import PATH_TO_EXPERIMENTS, QUALITATIVE_COLOR_PALETTE

experiment_ids = ['1674074113967956', '1674074113731734', '1673350499432390']
names = ["1", "2", "3"]
reward_developments = {}
reward_bands = {}

cosucc_developments = {}
cosucc_bands = {}

for i, id in enumerate(experiment_ids):
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "meta.json")) as f:
        meta = json.load(f)
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "progress.json")) as f:
        progress = json.load(f)
    with open(os.path.join("../../../", PATH_TO_EXPERIMENTS, str(id), "statistics.json")) as f:
        statistics = json.load(f)

    exp_name = names[i]
    reward_developments[exp_name] = progress["rewards"]["mean"]
    reward_bands[exp_name] = (
        np.array(progress["rewards"]["mean"]) - .2 * np.array(progress["rewards"]["stdev"]),
        np.array(progress["rewards"]["mean"]) + .2 * np.array(progress["rewards"]["stdev"])
    )

    cosucc_developments[exp_name] = statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"]
    cosucc_bands[exp_name] = (
        np.array(statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"])
        - .2 * np.array(statistics["auxiliary_performances"]["consecutive_goals_reached"]["std"]),

        np.array(statistics["auxiliary_performances"]["consecutive_goals_reached"]["mean"])
        + .2 * np.array(statistics["auxiliary_performances"]["consecutive_goals_reached"]["std"])
    )

fig, axes = plt.subplots(1, 2)

x_max = len(list(reward_developments.items())[0][1])
for i, (name, rewards) in enumerate(reward_developments.items()):
    x_max = min(x_max, len(rewards))
    axes[0].plot(rewards, label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    axes[0].fill_between(range(len(rewards)), reward_bands[name][0], reward_bands[name][1], alpha=0.1)

    axes[1].plot(cosucc_developments[name], label=name, color=QUALITATIVE_COLOR_PALETTE[i])
    axes[1].fill_between(range(len(rewards)), cosucc_bands[name][0], cosucc_bands[name][1], alpha=0.1)

axes[0].set_xlabel("Cycle")
axes[0].set_ylabel("Avg. Episode Return")
axes[0].legend()

axes[1].set_xlabel("Cycle")
axes[1].set_ylabel("Avg. Consecutive Goals Reached")
axes[1].legend()

axes[0].set_xlim(0, x_max)
axes[0].set_ylim(0)

axes[1].set_xlim(0, x_max)
axes[1].set_ylim(0)

plt.gcf().set_size_inches(16, 4)
plt.show()
# plt.savefig("../../../docs/figures/manipulate-shared-ablation.pdf", format="pdf", bbox_inches="tight")


import json
import os
from typing import List

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from scipy.signal import savgol_filter

from agent.ppo_agent import PPOAgent
from dexterity.design.common import QUALITATIVE_COLOR_PALETTE, PATH_TO_EXPERIMENTS

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

AGENT_ID =    1587117437
os.chdir("../../../")

with open(f"{PATH_TO_EXPERIMENTS}/{AGENT_ID}/progress.json", "r") as f:
    data = json.load(f)

with open(f"{PATH_TO_EXPERIMENTS}/{AGENT_ID}/meta.json", "r") as f:
    meta = json.load(f)

agent = PPOAgent.from_agent_state(AGENT_ID, "b", path_modifier="")

mean_rewards = data["rewards"]["mean"]
mean_rewards_smooth = savgol_filter(mean_rewards, 51, 3)
std_rewards = data["rewards"]["stdev"]

axs: List[Axes]
fig: Figure = plt.figure(figsize=(12, 4))
grid = plt.GridSpec(1, 3)

progression_ax = fig.add_subplot(grid[:2])
progression_ax.set_xlim(0, len(mean_rewards))

progression_ax.axhline(meta["environment"]["reward_threshold"], ls="--", color="grey") \
    if (meta["environment"]["reward_threshold"] != "None" and meta["environment"]["reward_threshold"] is not None) else None
progression_ax.plot(mean_rewards, color=QUALITATIVE_COLOR_PALETTE[0], alpha=.4, label="True Per Cycle Rewards")
progression_ax.plot(mean_rewards_smooth, color=QUALITATIVE_COLOR_PALETTE[0], label="Smoothed Per Cycle Rewards", lw=1)
progression_ax.legend(loc="lower right")

progression_ax.set_ylabel("Mean Episode Reward")
progression_ax.set_xlabel("Cycle")

# evaluate
stats, classes = agent.evaluate(100)
df = pd.DataFrame({
    "rewards": stats.episode_rewards,
    "finger": [['ff', 'mf', 'rf', 'lf', 'th'][c] for c in classes],
}, columns=["rewards", "finger"])

boxplot_ax = fig.add_subplot(grid[2])

box_ax = sns.boxplot(y=stats.episode_rewards, ax=boxplot_ax, color=QUALITATIVE_COLOR_PALETTE[0],
                     boxprops=dict(alpha=.4), )
sns.swarmplot(data=df, x=[""] * len(df), y="rewards", hue="finger", palette="Set1", ax=boxplot_ax)

box_ax.set(xlabel="Best Model Evaluation")

plt.ylabel("Episode Reward")

plt.subplots_adjust(wspace=.5, bottom=0.2)
plt.savefig("docs/figures/freereachingrelative_results.pdf", format="pdf", bbox_inches='tight')
plt.show()
